#include <string.h>

#include <epicsAssert.h>
#include <epicsEvent.h>
#include <epicsMutex.h>

#include <asynDriver.h>
#include <asynCommonSyncIO.h>
#include <asynOctet.h>
#include <asynOctetSyncIO.h>

#include "imcalibLib.h"
#include "imca-asyn-utils.h"

struct IpClientContext {
  const char *inputEom;
  int inputEomLength;
  const char *outputEom;
  int outputEomLength;
};

struct IpServerContext {
  const char *inputEom;
  int inputEomLength;
  const char *outputEom;
  int outputEomLength;
  void *registrarPvt;
  epicsMutexId lock;
  epicsEventId clientConnectedEvent;
  ELLLIST serverSessions;
};

struct AsynUserNode {
  ELLNODE node;
  asynUser *user;
};

static asynStatus connectToDevice(asynUser *);
static asynStatus disconnectFromDevice(asynUser *);
static struct IpClientContext *createIpClientContext(void);
static void destroyIpClientContext(struct IpClientContext *);
static struct IpServerContext *createIpServerContext(void);
static void destroyIpServerContext(struct IpServerContext *);
static struct AsynUserNode *createAsynUserNode(void);
static void destroyAsynUserNode(struct AsynUserNode *);

/*
 * The client asynUser obtained via this function must be destroyed via the
 * imcaDestroyIpClient function when no longer used.
 */
asynStatus imcaCreateConnectedIpClient(asynUser **client,
    const char *asynPortName, const char *inputEom, int inputEomLength,
    const char *outputEom, int outputEomLength) {
  asynUser *newClient;
  asynStatus status;

  status = imcaCreateIpClient(&newClient, asynPortName, inputEom,
    inputEomLength, outputEom, outputEomLength);
  if (status != asynSuccess) return status;

  status = imcaConnectIpClient(newClient);
  if (status != asynSuccess) {
    asynPrint(newClient, ASYN_TRACE_ERROR,
      "%s:error: in imcaCreateConnectedIpClient, imcaConnectIpClient failed "
      "(status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? newClient->errorMessage : "");
    imcaDestroyIpClient(newClient);
    return status;
  }

  *client = newClient;
  return asynSuccess;
}

/*
 * The client asynUser obtained via this function must be destroyed via the
 * imcaDestroyIpClient function when no longer used.
 */
asynStatus imcaCreateIpClient(asynUser **client, const char *asynPortName,
    const char *inputEom, int inputEomLength, const char *outputEom,
    int outputEomLength) {
  asynStatus status;
  asynUser *nullUser;
  struct IpClientContext *context;

  status = pasynOctetSyncIO->connect(asynPortName, 0, client, NULL);
  if (status != asynSuccess) {
    nullUser = pasynManager->createAsynUser(NULL, NULL);
    asynPrint(nullUser, ASYN_TRACE_ERROR, "%s:error: in imcaCreateIpClient, "
      "pasynOctetSyncIO->connect failed (status=%d) for asyn port \"%s\"\n",
      IMCALIB_NAME, status, asynPortName);
    pasynManager->freeAsynUser(nullUser);
    return status;
  }

  context = createIpClientContext();
  context->inputEom = inputEom;
  context->inputEomLength = inputEomLength;
  context->outputEom = outputEom;
  context->outputEomLength = outputEomLength;
  (*client)->userData = (void *)context;

  return asynSuccess;
}

asynStatus imcaConnectIpClient(asynUser *client) {
  struct IpClientContext *context;
  const char *asynPortName;
  asynStatus status;

  context = (struct IpClientContext *)client->userData;

  asynPortName = "";
  status = pasynManager->getPortName(client, &asynPortName);
  if (status != asynSuccess) asynPortName = "<unknown>";

  status = connectToDevice(client);
  if (status != asynSuccess) return status;

  status = pasynOctetSyncIO->setInputEos(client, context->inputEom,
    context->inputEomLength);
  if (status != asynSuccess) return status;

  return pasynOctetSyncIO->setOutputEos(client, context->outputEom,
    context->outputEomLength);
}

asynStatus imcaDisconnectIpClient(asynUser *client) {
  return disconnectFromDevice(client);
}

asynStatus imcaDestroyIpClient(asynUser *client) {
  struct IpClientContext *context;
  const char *asynPortName;
  asynStatus status;

  if (client == NULL) return asynSuccess;

  context = (struct IpClientContext *)client->userData;

  asynPortName = "";
  status = pasynManager->getPortName(client, &asynPortName);
  if (status != asynSuccess) asynPortName = "<unknown>";

  status = imcaDisconnectIpClient(client);
  if (status != asynSuccess) {
    asynPrint(client, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaDestroyIpClient, imcaDisconnectIpClient failed "
      "(status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? client->errorMessage : "");
    pasynOctetSyncIO->disconnect(client);
    destroyIpClientContext(context);
    return status;
  }

  status = pasynOctetSyncIO->disconnect(client);
  if (status != asynSuccess) {
    asynPrint(client, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaDestroyIpClient, pasynOctetSyncIO->disconnect failed "
      "(status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? client->errorMessage : "");
    destroyIpClientContext(context);
    return status;
  }

  destroyIpClientContext(context);

  return asynSuccess;
}

static void ipServerConnectCallback(void *drvPvt,
    asynUser *serverConnectionHandler, char *asynPortName, size_t len,
    int eomReason) {
  struct IpServerContext *context;
  asynUser *serverSession;
  asynUser *nullUser;
  struct AsynUserNode *serverSessionNode;
  asynStatus status;

  context = (struct IpServerContext *)drvPvt;

  status = pasynOctetSyncIO->connect(asynPortName, 0, &serverSession, NULL);
  if (status != asynSuccess) {
    nullUser = pasynManager->createAsynUser(NULL, NULL);
    asynPrint(nullUser, ASYN_TRACE_ERROR, "%s:error: in "
      "ipServerConnectCallback, pasynOctetSyncIO->connect failed (status=%d) "
      "for asyn port \"%s\"\n", IMCALIB_NAME, status, asynPortName);
    pasynManager->freeAsynUser(nullUser);
    return;
  }

  status = pasynOctetSyncIO->setInputEos(serverSession, context->inputEom,
    context->inputEomLength);
  if (status != asynSuccess) {
    asynPrint(serverSession, ASYN_TRACE_ERROR, "%s:error: in "
      "ipServerConnectCallback, pasynOctetSyncIO->setInputEos failed "
      "(status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? serverSession->errorMessage : "");
    pasynOctetSyncIO->disconnect(serverSession);
    return;
  }

  status = pasynOctetSyncIO->setOutputEos(serverSession, context->outputEom,
    context->outputEomLength);
  if (status != asynSuccess) {
    asynPrint(serverSession, ASYN_TRACE_ERROR, "%s:error: in "
      "ipServerConnectCallback, pasynOctetSyncIO->setOutputEos failed "
      "(status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? serverSession->errorMessage : "");
    pasynOctetSyncIO->disconnect(serverSession);
    return;
  }

  serverSessionNode = createAsynUserNode();
  serverSessionNode->user = serverSession;
  epicsMutexMustLock(context->lock);
  ellAdd(&context->serverSessions, (ELLNODE *)serverSessionNode);
  epicsMutexUnlock(context->lock);
  epicsEventSignal(context->clientConnectedEvent);
}

/*
 * The server connection handler asynUser obtained via this function must be
 * destroyed via the imcaDestroyIpServer function when no longer used.
 */
asynStatus imcaCreateListeningIpServer(
    asynUser **serverConnectionHandler, const char *asynPortName,
    const char *inputEom, int inputEomLength, const char *outputEom,
    int outputEomLength) {
  asynUser *connectionHandler;
  struct IpServerContext *context;
  asynStatus status;
  asynInterface *pasynInterface;
  asynOctet *pasynOctet;

  connectionHandler = pasynManager->createAsynUser(NULL, NULL);
  context = createIpServerContext();
  context->inputEom = inputEom;
  context->inputEomLength = inputEomLength;
  context->outputEom = outputEom;
  context->outputEomLength = outputEomLength;
  connectionHandler->userData = (void *)context;

  status = pasynManager->connectDevice(connectionHandler, asynPortName, 0);
  if (status != asynSuccess) {
    asynPrint(connectionHandler, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaCreateListeningIpServer, pasynManager->connectDevice failed "
      "(status=%d) while trying to connect to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? connectionHandler->errorMessage : "");
    destroyIpServerContext(context);
    pasynManager->freeAsynUser(connectionHandler);
    return status;
  }

  pasynInterface = pasynManager->findInterface(connectionHandler,
    asynOctetType, 1);
  if (pasynInterface == NULL) {
    asynPrint(connectionHandler, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaCreateListeningIpServer, pasynManager->findInterface failed to "
      "find interface \"%s\" for asyn user connected to port \"%s\"\n",
      IMCALIB_NAME, asynOctetType, asynPortName);
    destroyIpServerContext(context);
    pasynManager->freeAsynUser(connectionHandler);
    return asynError;
  }

  pasynOctet = (asynOctet *)pasynInterface->pinterface;
  status = pasynOctet->registerInterruptUser(pasynInterface->drvPvt,
    connectionHandler, ipServerConnectCallback, context,
    &context->registrarPvt);
  if (status != asynSuccess) {
    asynPrint(connectionHandler, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaCreateListeningIpServer, pasynOctet->registerInterruptUser "
      "failed (status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? connectionHandler->errorMessage : "");
    destroyIpServerContext(context);
    pasynManager->freeAsynUser(connectionHandler);
    return status;
  }

  *serverConnectionHandler = connectionHandler;
  return asynSuccess;
}

/*
 * The server session asynUser obtained via this function must be destroyed
 * via the imcaDestroyIpServerSession function when no longer used.
 */
epicsEventWaitStatus imcaWaitForIpServerConnection(asynUser **serverSession,
    asynUser *serverConnectionHandler, double timeout) {
  struct IpServerContext *context;
  int waitedOnce;
  epicsEventWaitStatus waitStatus;
  struct AsynUserNode *node;

  context = (struct IpServerContext *)serverConnectionHandler->userData;

  waitedOnce = 0;
  waitStatus = epicsEventWaitError;
  do {
    epicsMutexMustLock(context->lock);
    node = (struct AsynUserNode *)ellGet(&context->serverSessions);
    epicsMutexUnlock(context->lock);
    if (node != NULL) {
      *serverSession = node->user;
      destroyAsynUserNode(node);
      return epicsEventWaitOK;
    }
    if (waitedOnce) return epicsEventWaitError;
    waitStatus = epicsEventWaitWithTimeout(context->clientConnectedEvent,
      timeout);
    waitedOnce = 1;
  } while (waitStatus == epicsEventWaitOK);

  return waitStatus;
}

asynStatus imcaDestroyIpServerSession(asynUser *serverSession) {
  const char *asynPortName;
  asynStatus status;

  if (serverSession == NULL) return asynSuccess;

  asynPortName = "";
  status = pasynManager->getPortName(serverSession, &asynPortName);
  if (status != asynSuccess) asynPortName = "<unknown>";

  status = disconnectFromDevice(serverSession);
  if (status != asynSuccess) {
    asynPrint(serverSession, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaDestroyIpServerSession, disconnectFromDevice failed (status=%d) "
      "for asyn user connected to port \"%s\"%s%s\n", IMCALIB_NAME, status,
      asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? serverSession->errorMessage : "");
    pasynOctetSyncIO->disconnect(serverSession);
    return status;
  }

  status = pasynOctetSyncIO->disconnect(serverSession);
  if (status != asynSuccess) {
    asynPrint(serverSession, ASYN_TRACE_ERROR, "%s:error: in "
      "imcaDestroyIpServerSession, pasynOctetSyncIO->disconnect failed "
      "(status=%d) for asyn user connected to port \"%s\"%s%s\n",
      IMCALIB_NAME, status, asynPortName, (status == asynError) ? ": " : "",
      (status == asynError) ? serverSession->errorMessage : "");
  }
  return status;
}

asynStatus imcaDestroyIpServer(asynUser *serverConnectionHandler) {
  struct IpServerContext *context;
  asynInterface *pasynInterface;
  asynOctet *pasynOctet;
  asynStatus status;

  if (serverConnectionHandler == NULL) return asynSuccess;

  context = (struct IpServerContext *)serverConnectionHandler->userData;

  pasynInterface = pasynManager->findInterface(serverConnectionHandler,
    asynOctetType, 1);
  if (pasynInterface == 0) return asynError;

  pasynOctet = (asynOctet *)pasynInterface->pinterface;
  status = pasynOctet->cancelInterruptUser(pasynInterface->drvPvt,
    serverConnectionHandler, context->registrarPvt);
  if (status != asynSuccess) return status;

  status = pasynManager->freeAsynUser(serverConnectionHandler);
  if (status != asynSuccess) return status;

  destroyIpServerContext(context);

  return asynSuccess;
}

static asynStatus connectToDevice(asynUser *user) {
  const char *asynPortName;
  asynStatus status;
  int autoConnect;
  asynUser *commonUser;
  int connected;

  asynPortName = "";
  status = pasynManager->getPortName(user, &asynPortName);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:connectToDevice, pasynManager->getPortName failed (status=%d)",
      IMCALIB_NAME, status);
    return asynError;
  }

  autoConnect = 1;
  status = pasynManager->isAutoConnect(user, &autoConnect);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:connectToDevice, pasynManager->isAutoConnect failed (status=%d)",
      IMCALIB_NAME, status);
    return asynError;
  }
  if (autoConnect) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:connectToDevice, asynUser has auto-connect enabled, but "
      "auto-connect should be disabled when used with this function",
      IMCALIB_NAME);
    return asynError;
  }

  status = pasynCommonSyncIO->connect(asynPortName, 0, &commonUser, NULL);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:connectToDevice, pasynCommonSyncIO->connect failed (status=%d)",
      IMCALIB_NAME, status);
    return asynError;
  }

  connected = 0;
  status = pasynManager->isConnected(commonUser, &connected);
  if (status != asynSuccess) connected = 0;

  if (connected) {
    status = pasynCommonSyncIO->disconnect(commonUser);
    if (status == asynSuccess) return status;
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:connectToDevice, pasynCommonSyncIO->disconnect failed "
      "(status=%d)%s%s", IMCALIB_NAME, status,
      (status == asynError) ? ": " : "",
      (status == asynError) ? commonUser->errorMessage : "");
    return asynError;
  }

  status = pasynCommonSyncIO->connectDevice(commonUser);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:connectToDevice, pasynCommonSyncIO->connectDevice failed "
      "(status=%d)%s%s", IMCALIB_NAME, status,
      (status == asynError) ? ": " : "",
      (status == asynError) ? commonUser->errorMessage : "");
    pasynCommonSyncIO->disconnect(commonUser);
    return asynError;
  }

  status = pasynCommonSyncIO->disconnect(commonUser);
  if (status == asynSuccess) return status;
  epicsSnprintf(user->errorMessage, user->errorMessageSize,
    "in %s:connectToDevice, pasynCommonSyncIO->disconnect failed "
    "(status=%d)%s%s", IMCALIB_NAME, status,
    (status == asynError) ? ": " : "",
    (status == asynError) ? commonUser->errorMessage : "");
  return asynError;
}

static asynStatus disconnectFromDevice(asynUser *user) {
  const char *asynPortName;
  asynStatus status;
  asynUser *commonUser;
  int connected;

  if (user == NULL) return asynSuccess;

  asynPortName = "";
  status = pasynManager->getPortName(user, &asynPortName);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:disconnectFromDevice, pasynManager->getPortName failed "
      "(status=%d)", IMCALIB_NAME, status);
    return asynError;
  }

  status = pasynCommonSyncIO->connect(asynPortName, 0, &commonUser, NULL);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:disconnectFromDevice, pasynCommonSyncIO->connect failed "
      "(status=%d)", IMCALIB_NAME, status);
    return asynError;
  }

  connected = 1;
  status = pasynManager->isConnected(commonUser, &connected);
  if (status != asynSuccess) connected = 1;

  if (!connected) {
    status = pasynCommonSyncIO->disconnect(commonUser);
    if (status == asynSuccess) return status;
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:disconnectFromDevice, pasynCommonSyncIO->disconnect failed "
      "(status=%d)%s%s", IMCALIB_NAME, status,
      (status == asynError) ? ": " : "",
      (status == asynError) ? commonUser->errorMessage : "");
    return asynError;
  }

  status = pasynCommonSyncIO->disconnectDevice(commonUser);
  if (status != asynSuccess) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "in %s:disconnectFromDevice, pasynCommonSyncIO->disconnectDevice "
      "failed (status=%d)%s%s", IMCALIB_NAME, status,
      (status == asynError) ? ": " : "",
      (status == asynError) ? commonUser->errorMessage : "");
    pasynCommonSyncIO->disconnect(commonUser);
    return asynError;
  }

  status = pasynCommonSyncIO->disconnect(commonUser);
  if (status == asynSuccess) return status;
  epicsSnprintf(user->errorMessage, user->errorMessageSize,
    "in %s:disconnectFromDevice pasynCommonSyncIO->disconnect failed "
    "(status=%d)%s%s", IMCALIB_NAME, status,
    (status == asynError) ? ": " : "",
    (status == asynError) ? commonUser->errorMessage : "");
  return asynError;
}

static struct IpClientContext *createIpClientContext(void) {
  struct IpClientContext *context;

  context = NULL;
  context = (struct IpClientContext *)pasynManager->memMalloc(sizeof(
    struct IpClientContext));
  assert(context != NULL);
  context->inputEom = "";
  context->inputEomLength = 0;
  context->outputEom = "";
  context->outputEomLength = 0;
  return context;
}

static void destroyIpClientContext(struct IpClientContext *context) {
  if (context == NULL) return;
  pasynManager->memFree(context, sizeof(struct IpClientContext));
}

static struct IpServerContext *createIpServerContext(void) {
  struct IpServerContext *context;

  context = NULL;
  context = (struct IpServerContext *)pasynManager->memMalloc(sizeof(
    struct IpServerContext));
  assert(context != NULL);
  context->inputEom = "";
  context->inputEomLength = 0;
  context->outputEom = "";
  context->outputEomLength = 0;
  context->registrarPvt = NULL;
  context->lock = epicsMutexMustCreate();
  context->clientConnectedEvent = epicsEventMustCreate(epicsEventEmpty);
  ellInit(&context->serverSessions);
  return context;
}

static void destroyIpServerContext(struct IpServerContext *context) {
  struct AsynUserNode *node;

  if (context == NULL) return;
  while (ellCount(&context->serverSessions) > 0) {
    node = (struct AsynUserNode *)ellGet(&context->serverSessions);
    imcaDestroyIpServerSession(node->user);
    destroyAsynUserNode(node);
  }
  epicsEventDestroy(context->clientConnectedEvent);
  epicsMutexDestroy(context->lock);
  pasynManager->memFree(context, sizeof(struct IpServerContext));
}

static struct AsynUserNode *createAsynUserNode(void) {
  struct AsynUserNode *node;

  node = NULL;
  node = (struct AsynUserNode *)pasynManager->memMalloc(sizeof(
    struct AsynUserNode));
  assert(node != NULL);
  return node;
}

static void destroyAsynUserNode(struct AsynUserNode *node) {
  if (node == NULL) return;
  pasynManager->memFree(node, sizeof(struct AsynUserNode));
}

/* value must be null-terminated string */
asynStatus imcaAsynWriteString(asynUser *user, const char *value,
    double timeout) {
  return imcaAsynWrite(user, value, strlen(value), timeout);
}

asynStatus imcaAsynReadString(asynUser *user, char *value,
    size_t valueSize, double timeout) {
  asynStatus status;
  size_t numBytesTransferred;
  int eomReason;

  if (valueSize <= 0) return asynSuccess;
  value[0] = '\0';
  status = pasynOctetSyncIO->read(user, value, valueSize - 1, timeout,
    &numBytesTransferred, &eomReason);
  if (status != asynSuccess) return status;
  if ((eomReason & ASYN_EOM_EOS) != ASYN_EOM_EOS) {
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "only part of string was transferred (number of bytes transferred = "
      "%d)", (int)numBytesTransferred);
    return asynError;
  }
  value[numBytesTransferred] = '\0';
  return asynSuccess;
}

asynStatus imcaAsynBufferedWrite(asynUser *user, char *buffer,
    size_t bufferSize, int *bufferLength, char whatToWrite, double timeout) {
  asynStatus status;

  if (bufferSize < 1) return asynError;

  if (*bufferLength >= bufferSize) {
    status = imcaAsynWrite(user, buffer, *bufferLength, timeout);
    if (status != asynSuccess) return status;
    *bufferLength = 0;
  }

  buffer[*bufferLength] = whatToWrite;
  (*bufferLength)++;
  return asynSuccess;
}

asynStatus imcaAsynWrite(asynUser *user, const char *whatToWrite,
    size_t whatToWriteLength, double timeout) {
  size_t numBytesTransferred;
  asynStatus status;

  numBytesTransferred = 0;
  status = pasynOctetSyncIO->write(user, whatToWrite, whatToWriteLength,
    timeout, &numBytesTransferred);
  if (status != asynSuccess) return status;
  if (numBytesTransferred != whatToWriteLength) {
    status = asynError;
    epicsSnprintf(user->errorMessage, user->errorMessageSize,
      "only part of whatToWrite was transferred (number of bytes transferred "
      "= %d, expected = %d)", (int)numBytesTransferred,
     (int)whatToWriteLength);
  }
  return status;
}
