#include <math.h>
#include <stdlib.h>
#include <string.h>

#include <cadef.h>
#include <db_access.h>
#include <epicsAssert.h>

#include "imca-ca-utils.h"

#define WAIT_FOR_PV_TO_HAVE_VALUE_DELTA 1e-15
#define WAIT_FOR_PV_TO_HAVE_VALUE_TIMEOUT 30.0
#define WAIT_FOR_PV_TO_HAVE_VALUE_POLL_DELAY 0.25

static int areEpicsValuesEqual(chtype, void *, void *, double);

int imcaSetEpicsPvToChar(const char *pvName, dbr_char_t value) {
  chid pvChannel;
  int status;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_put(DBR_CHAR, pvChannel, &value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaGetCharFromEpicsPv(const char *pvName, dbr_char_t *value) {
  chid pvChannel;
  int status;

  *value = '\0';

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_get(DBR_CHAR, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaSetEpicsPvToDouble(const char *pvName, dbr_double_t value) {
  chid pvChannel;
  int status;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_put(DBR_DOUBLE, pvChannel, &value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaGetDoubleFromEpicsPv(const char *pvName, dbr_double_t *value) {
  chid pvChannel;
  int status;

  *value = 0.0;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_get(DBR_DOUBLE, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaSetEpicsPvToEnum(const char *pvName, dbr_enum_t value) {
  chid pvChannel;
  int status;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_put(DBR_ENUM, pvChannel, &value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaGetEnumFromEpicsPv(const char *pvName, dbr_enum_t *value) {
  chid pvChannel;
  int status;

  *value = 0;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_get(DBR_ENUM, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaSetEpicsPvToLong(const char *pvName, dbr_long_t value) {
  chid pvChannel;
  int status;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_put(DBR_LONG, pvChannel, &value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaGetLongFromEpicsPv(const char *pvName, dbr_long_t *value) {
  chid pvChannel;
  int status;

  *value = 0;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_get(DBR_LONG, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaSetEpicsPvToShort(const char *pvName, dbr_short_t value) {
  chid pvChannel;
  int status;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_put(DBR_SHORT, pvChannel, &value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaGetShortFromEpicsPv(const char *pvName, dbr_short_t *value) {
  chid pvChannel;
  int status;

  *value = 0;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_get(DBR_SHORT, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaSetEpicsPvToString(const char *pvName, dbr_string_t value) {
  chid pvChannel;
  int status;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_array_put(DBR_STRING, 1, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaGetStringFromEpicsPv(const char *pvName, dbr_string_t value) {
  chid pvChannel;
  int status;

  value[0] = '\0';

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  status = ca_array_get(DBR_STRING, 1, pvChannel, value);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  return ECA_NORMAL;
}

int imcaWaitForEpicsPvToHaveValue(const char *pvName, chtype type,
    void *value) {
  return imcaWaitForEpicsPvToHaveValueWithDeltaAndTimeout(pvName, type, value,
    WAIT_FOR_PV_TO_HAVE_VALUE_DELTA, WAIT_FOR_PV_TO_HAVE_VALUE_TIMEOUT);
}

int imcaWaitForEpicsPvToHaveValueWithDelta(const char *pvName, chtype type,
    void *value, double delta) {
  return imcaWaitForEpicsPvToHaveValueWithDeltaAndTimeout(pvName, type, value,
    delta, WAIT_FOR_PV_TO_HAVE_VALUE_TIMEOUT);
}

int imcaWaitForEpicsPvToHaveValueWithTimeout(const char *pvName, chtype type,
    void *value, double timeout) {
  return imcaWaitForEpicsPvToHaveValueWithDeltaAndTimeout(pvName, type, value,
    WAIT_FOR_PV_TO_HAVE_VALUE_DELTA, timeout);
}

int imcaWaitForEpicsPvToHaveValueWithDeltaAndTimeout(const char *pvName,
    chtype type, void *value, double delta, double timeout) {
  chid pvChannel;
  int status;
  void *fetchedValue;
  epicsTimeStamp startTime, endTime;
  double elapsedTime;
  int foundMatch;

  status = ca_context_create(ca_enable_preemptive_callback);
  if (status != ECA_NORMAL) {
    return status;
  }

  status = ca_create_channel(pvName, NULL, NULL, CA_PRIORITY_DEFAULT,
    &pvChannel);
  if (status != ECA_NORMAL) {
    ca_context_destroy();
    return status;
  }
  status = ca_pend_io(5.0);
  if (status != ECA_NORMAL) {
    ca_clear_channel(pvChannel);
    ca_context_destroy();
    return status;
  }

  foundMatch = 0;
  fetchedValue = malloc(dbr_size_n(type, 1));
  assert(fetchedValue != NULL);
  status = epicsTimeGetCurrent(&startTime);
  assert(status == epicsTimeOK);
  for (;;) {
    status = ca_get(type, pvChannel, fetchedValue);
    if (status != ECA_NORMAL) {
      ca_clear_channel(pvChannel);
      ca_context_destroy();
      free(fetchedValue);
      return status;
    }
    status = ca_pend_io(5.0);
    if (status != ECA_NORMAL) {
      ca_clear_channel(pvChannel);
      ca_context_destroy();
      free(fetchedValue);
      return status;
    }
    if (areEpicsValuesEqual(type, value, fetchedValue, delta)) {
      foundMatch = 1;
      break;
    }
    status = epicsTimeGetCurrent(&endTime);
    assert(status == epicsTimeOK);
    elapsedTime = epicsTimeDiffInSeconds(&endTime, &startTime);
    if (elapsedTime > timeout) break;
    epicsThreadSleep(WAIT_FOR_PV_TO_HAVE_VALUE_POLL_DELAY);
  }

  ca_clear_channel(pvChannel);
  ca_context_destroy();
  free(fetchedValue);

  return foundMatch ? ECA_NORMAL : ECA_TIMEOUT;
}

static int areEpicsValuesEqual(chtype type, void *val1, void *val2,
    double delta) {
  if (type == DBR_CHAR)
    return *((dbr_char_t *)val1) == *((dbr_char_t *)val2);
  if (type == DBR_SHORT)
    return *((dbr_short_t *)val1) == *((dbr_short_t *)val2);
  if (type == DBR_ENUM)
    return *((dbr_enum_t *)val1) == *((dbr_enum_t *)val2);
  if (type == DBR_LONG)
    return *((dbr_long_t *)val1) == *((dbr_long_t *)val2);
  if (type == DBR_FLOAT)
    return fabsf(*((dbr_float_t *)val1) - *((dbr_float_t *)val2)) <= delta;
  if (type == DBR_DOUBLE)
    return fabs(*((dbr_double_t *)val1) - *((dbr_double_t *)val2)) <= delta;
  if (type == DBR_STRING)
    return strcmp(*((dbr_string_t *)val1), *((dbr_string_t *)val2)) == 0;
  assert(1); /* unsupported value type */
  return 0; /* unreached */
}
