#include <math.h>
#include <time.h>
#include <string.h>
#include <errlog.h>
#include <epicsAssert.h>
#include "imca-epics-unit.h"

void imcaFail(EpicsUnitContext *context, const char *fileName, int lineNum) {
  if (context->testThatFailed != context->testCounter) {
    context->testThatFailed = context->testCounter;
    context->numTestsFailed++;
  }
  errlogFlush();
  errlogPrintf("\nTest case failed at %s:%d\n", fileName, lineNum);
  errlogFlush();
}

void imcaFailIfIntNotEqual(EpicsUnitContext *context, int value1, int value2,
    const char *fileName, int lineNum) {
  if (value1 == value2) return;
  imcaFail(context, fileName, lineNum);
  errlogPrintf("because %d != %d\n", value1, value2);
  errlogFlush();
}

void imcaFailIfDoubleNotEqual(EpicsUnitContext *context, double value1,
    double value2, double delta, const char *fileName, int lineNum) {
  if (delta == 0.0 && value1 == value2) return;
  if (fabs(value1 - value2) <= delta) return;
  imcaFail(context, fileName, lineNum);
  errlogPrintf("because %.9f != %.9f\n", value1, value2);
  errlogFlush();
}

void imcaFailIfDoubleArrayNotEqual(EpicsUnitContext *context,
    double values1[], double values2[], size_t valuesSize, double delta,
    const char *fileName, int lineNum) {
  int i;

  for (i = 0; i < valuesSize; i++) {
    if (delta == 0.0 && values1[i] == values2[i]) continue;
    if (fabs(values1[i] - values2[i]) <= delta) continue;
    imcaFail(context, fileName, lineNum);
    errlogPrintf("because %.9f != %.9f at index %d\n", values1[i], values2[i],
      i);
    errlogFlush();
    return;
  }
}

void imcaFailIfStringNotEqual(EpicsUnitContext *context, const char *value1,
    const char *value2, const char *fileName, int lineNum) {
  if (value1 == value2) return;
  if (value1 != NULL && value2 != NULL && strcmp(value1, value2) == 0) return;
  imcaFail(context, fileName, lineNum);
  errlogPrintf("because \"%s\" != \"%s\"\n", value1, value2);
  errlogFlush();
}

void imcaNoteNowAsEpicsUnitTestsStartTime(EpicsUnitContext *context) {
  time(&context->testsStartTime);
  assert(context->testsStartTime != (time_t)-1);
}

void imcaNoteNowAsEpicsUnitTestsEndTime(EpicsUnitContext *context) {
  time(&context->testsEndTime);
  assert(context->testsEndTime != (time_t)-1);
}

void imcaPrintEpicsUnitTestProgress(void) {
  errlogFlush();
  errlogMessage(".");
  errlogFlush();
}

void imcaPrintEpicsUnitTestsStarting(void) {
  errlogFlush();
  errlogMessage("Running tests ");
  errlogFlush();
}

void imcaPrintEpicsUnitTestsSummary(EpicsUnitContext *context) {
  double runTimeInSeconds;
  double runTimeMinutesPart;
  double runTimeSecondsPart;

  errlogFlush();
  errlogMessage("\n");
  if (context->numTestsFailed > 0) {
    errlogPrintf("FAILURE: %d test%s failed.\n", context->numTestsFailed,
      context->numTestsFailed == 1 ? "" : "s");
  } else {
    errlogMessage("SUCCESS: All tests passed.\n");
  }
  errlogPrintf("Tests: %d.\n", context->testCounter);
  if (context->testsStartTime != -1 && context->testsEndTime != -1) {
    runTimeInSeconds = difftime(context->testsEndTime,
      context->testsStartTime);
    modf(runTimeInSeconds / 60.0, &runTimeMinutesPart);
    runTimeSecondsPart = fmod(runTimeInSeconds, 60.0);
    errlogPrintf("Time: %.0fm%.0fs.\n", runTimeMinutesPart,
      runTimeSecondsPart);
  }
  errlogFlush();
}

void imcaInitializeEpicsUnit(EpicsUnitContext *context) {
  context->testThatFailed = 0;
  context->numTestsFailed = 0;
  context->testCounter = 0;
  context->testsStartTime = (time_t)-1;
  context->testsEndTime = (time_t)-1;
}

void imcaStartingEpicsUnitTest(EpicsUnitContext *context) {
  context->testCounter++;
  imcaPrintEpicsUnitTestProgress();
}
