#include <stdarg.h>
#include <epicsAssert.h>
#include <epicsStdio.h>
#include <epicsTime.h>
#include <errlog.h>
#include "imca-errlog-utils.h"

#define TIMESTAMP_SIZE 24
#define FORMATTED_WITH_TIMESTAMP_SIZE 256

static int atomicErrlogTsVprintf(const char *pformat, va_list pvar);
static int nonatomicErrlogTsVprintf(const char *pformat, va_list pvar);
static int printTimestamp(char *buffer, size_t bufferSize);

int imcaErrlogTsPrintf(const char *pformat, ...) {
  va_list pvar;
  int numCharsPrinted;

  va_start(pvar, pformat);
  numCharsPrinted = atomicErrlogTsVprintf(pformat, pvar);
  va_end(pvar);
  if (numCharsPrinted != -1) return numCharsPrinted;

  va_start(pvar, pformat);
  numCharsPrinted = nonatomicErrlogTsVprintf(pformat, pvar);
  va_end(pvar);
  return numCharsPrinted;
}

int imcaErrlogTsVprintf(const char *pformat, va_list pvar) {
  return nonatomicErrlogTsVprintf(pformat, pvar);
}

static int atomicErrlogTsVprintf(const char *pformat, va_list pvar) {
  char formattedWithTimestamp[FORMATTED_WITH_TIMESTAMP_SIZE];
  int numCharsPrinted;

  numCharsPrinted = printTimestamp(formattedWithTimestamp,
    sizeof(formattedWithTimestamp));
  assert(numCharsPrinted == TIMESTAMP_SIZE - 1);
  assert(numCharsPrinted < FORMATTED_WITH_TIMESTAMP_SIZE - 2);
  numCharsPrinted += epicsSnprintf(formattedWithTimestamp + numCharsPrinted,
    sizeof(formattedWithTimestamp) - numCharsPrinted, " ");
  numCharsPrinted += epicsVsnprintf(formattedWithTimestamp + numCharsPrinted,
    sizeof(formattedWithTimestamp) - numCharsPrinted, pformat, pvar);
  if (numCharsPrinted >= FORMATTED_WITH_TIMESTAMP_SIZE) return -1;
  return errlogPrintf("%s", formattedWithTimestamp);
}

static int nonatomicErrlogTsVprintf(const char *pformat, va_list pvar) {
  char timestamp[TIMESTAMP_SIZE];
  int numCharsPrinted;

  numCharsPrinted = printTimestamp(timestamp, sizeof(timestamp));
  assert(numCharsPrinted == TIMESTAMP_SIZE - 1);
  numCharsPrinted = errlogPrintf("%s ", timestamp);
  numCharsPrinted += errlogVprintf(pformat, pvar);
  return numCharsPrinted;
}

static int printTimestamp(char *buffer, size_t bufferSize) {
  int status;
  epicsTimeStamp timestampNow;
  size_t numCharsPrinted;

  if (bufferSize <= 0) return 0;
  status = epicsTimeGetCurrent(&timestampNow);
  if (status != epicsTimeOK) return 0;
  numCharsPrinted = epicsTimeToStrftime(buffer, bufferSize,
    "%Y/%m/%d %H:%M:%S.%03f", &timestampNow);
  buffer[numCharsPrinted] = '\0';
  return numCharsPrinted;
}
