#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <epicsAssert.h>
#include <epicsStdio.h>
#include <asynDriver.h>
#include <asynOctet.h>
#include <asynOctetSyncIO.h>
#include "imca-asyn-utils.h"
#include "imca-string-utils.h"
#include "imca-smtp-client.h"

/* Public interface */

/*
 * Sends an email message via SMTP.
 *
 * The environment variable IMCA_SMTP_HELO_HOST must be set to the string that
 * will be used as the host parameter to the SMTP HELO request that will be
 * sent to the SMTP server. This must be the correct hostname for the computer
 * calling this function that is acting as the SMTP client. For a computer
 * that does not have a hostname, the IP address of the interface that will be
 * used to communicate with the SMTP server should be used inside square
 * brackets (e.g. "[164.54.200.67]").
 *
 * Considers the end of line for SMTP server responses to be \012 since some
 * SMTP servers do not correctly send \015 before \012
 * (http://cr.yp.to/smtp/request.html#response).
 *
 * asynSmtpClientPortName
 *              Name of asyn TCP/IP port for communicating with SMTP server.
 *              Should be configured with auto-connect turned off (e.g.
 *              drvAsynIPPortConfigure("smtpClient","127.0.0.1:51034",0,1,0)).
 * from         Sender email address (SMTP envelope sender address). Must be a
 *              properly encoded RFC 822 email address (just the address).
 *              Must not be empty.
 * recipients   List of recipient addresses (SMTP envelope recipient
 *              addresses). All addresses must be properly encoded RFC 822
 *              email addresses (just the address). Must not be empty. Empty
 *              recipient addresses will be silently skipped.
 * message      Email message to send. Message must not contain lines longer
 *              than 997 bytes (996 bytes if line starts with a '.') such that
 *              after being encoded, the total line length is less than or
 *              equal to 998 bytes (RFC 821). Must not contain partial lines
 *              (i.e. a line that does not end with \012). Must not contain
 *              the byte sequence \015\012. Newline byte sequence must be
 *              \012.
 * smtpTimeout  Timeout for communication with SMTP server (-1.0 to use
 *              default timeout).
 *
 * Returns ImcaEmailStatusSuccess on sucess.
 */
ImcaEmailStatus imcaSendEmail(const char *asynSmtpClientPortName,
    const char *from, char *recipients[], int recipientsSize,
    const char *message, double smtpTimeout) {
  asynStatus status;
  asynUser *smtpClient;
  ImcaEmailStatus iestatus;

  status = imcaCreateConnectedIpClient(&smtpClient, asynSmtpClientPortName,
    "\n", 1, "", 0);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;

  iestatus = imcaSendEmailUsingAsynUser(smtpClient, from, recipients,
    recipientsSize, message, smtpTimeout);

  status = imcaDestroyIpClient(smtpClient);
  if (iestatus != ImcaEmailStatusSuccess) return iestatus;
  return status == asynSuccess ?
    ImcaEmailStatusSuccess :
    ImcaEmailStatusNetworkError;
}

/*
 * Writes an RFC 822 email message to dst.
 *
 * All (char *) values must be null-terminated strings.
 *
 * dst      String where the email message will be written.
 * dstSize  Size of dst string buffer.
 * from     From field value. Must not be empty.
 * to       To recipients. Must contain at least one recipient. Empty
 *          recipient addresses will be silently skipped.
 * toSize   Size of to array.
 * cc       Cc recipients. May be empty or NULL.
 * ccSize   Size of cc array.
 * subject  Subject field value. May be empty.
 * body     Body of email. Must use \012 (not \015\012 or other) as the
 *          newline sequence.
 *
 * Returns ImcaEmailStatusSuccess on sucess.
 */
ImcaEmailStatus imcaCreateEmailMessage(char *dst, size_t dstSize,
    const char *from, char *to[], int toSize, char *cc[], int ccSize,
    const char *subject, const char *body) {
  int count;
  int totalCount;
  size_t currentSize;
  time_t nowAsSeconds;
  struct tm nowAsTm;

  dst[0] = '\0';
  count = 0;
  totalCount = 0;
  currentSize = dstSize;
  time(&nowAsSeconds); assert(nowAsSeconds != (time_t)-1);
  gmtime_r(&nowAsSeconds, &nowAsTm);

  count = imcaPrintEmailDateField(&dst[totalCount], currentSize, &nowAsTm);
  if (count == -1) return ImcaEmailStatusBufferTooSmall;
  totalCount += count;
  currentSize -= count;

  count = imcaPrintEmailFromField(&dst[totalCount], currentSize, from);
  if (count == -1) return ImcaEmailStatusBufferTooSmall;
  if (count == -2) return ImcaEmailStatusEmptyFrom;
  totalCount += count;
  currentSize -= count;

  count = imcaPrintEmailToField(&dst[totalCount], currentSize, to, toSize);
  if (count == -1) return ImcaEmailStatusBufferTooSmall;
  if (count == -2) return ImcaEmailStatusEmptyTo;
  totalCount += count;
  currentSize -= count;

  count = imcaPrintEmailCcField(&dst[totalCount], currentSize, cc, ccSize);
  if (count == -1) return ImcaEmailStatusBufferTooSmall;
  totalCount += count;
  currentSize -= count;

  count = imcaPrintEmailSubjectField(&dst[totalCount], currentSize, subject);
  if (count == -1) return ImcaEmailStatusBufferTooSmall;
  totalCount += count;
  currentSize -= count;

  count = imcaPrintEmailBody(&dst[totalCount], currentSize, body);
  if (count == -1) return ImcaEmailStatusBufferTooSmall;
  totalCount += count;
  currentSize -= count;

  return ImcaEmailStatusSuccess;
}

/* Private interface */

/* Returns ImcaEmailStatusSuccess on sucess.
 */
ImcaEmailStatus imcaSendEmailUsingAsynUser(asynUser *smtpClient,
    const char *from, char *recipients[], int recipientsSize,
    const char *message, double smtpTimeout) {
  double timeout;
  ImcaEmailStatus status;

  timeout = smtpTimeout < 0.0 ? IMCA_SMTP_DEFAULT_TIMEOUT : smtpTimeout;

  status = imcaWaitForSmtpServerGreeting(smtpClient, timeout);
  if (status != ImcaEmailStatusSmtpSuccess) {
    imcaSendSmtpQuitRequest(smtpClient, timeout);
    return status;
  }

  status = imcaSendSmtpHeloRequest(smtpClient, timeout);
  if (status != ImcaEmailStatusSmtpSuccess) {
    imcaSendSmtpQuitRequest(smtpClient, timeout);
    return status;
  }

  status = imcaSendSmtpMailRequest(smtpClient, timeout, from);
  if (status != ImcaEmailStatusSmtpSuccess) {
    imcaSendSmtpQuitRequest(smtpClient, timeout);
    return status;
  }

  status = imcaSendSmtpRcptRequests(smtpClient, timeout, recipients,
    recipientsSize);
  if (status != ImcaEmailStatusSmtpSuccess) {
    imcaSendSmtpQuitRequest(smtpClient, timeout);
    return status;
  }

  status = imcaSendSmtpDataRequest(smtpClient, timeout);
  if (status != ImcaEmailStatusSmtpSuccess) {
    imcaSendSmtpQuitRequest(smtpClient, timeout);
    return status;
  }

  status = imcaSendSmtpEncodedMessage(smtpClient, timeout, message);
  if (status != ImcaEmailStatusSmtpSuccess) {
    if (status != ImcaEmailStatusNetworkError)
      imcaSendSmtpQuitRequest(smtpClient, timeout);
    return status;
  }

  status = imcaSendSmtpQuitRequest(smtpClient, timeout);
  return status == ImcaEmailStatusSmtpSuccess ?
    ImcaEmailStatusSuccess :
    status;
}

/* Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaWaitForSmtpServerGreeting(asynUser *client,
    double timeout) {
  return imcaReadSmtpResponse(client, timeout);
}

/* Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpHeloRequest(asynUser *client, double timeout) {
  char *heloHost;
  char request[IMCA_SMTP_REQUEST_SIZE + 1];
  int length;

  heloHost = imcaGetSmtpHeloHost();
  if (heloHost == NULL) return ImcaEmailStatusSmtpHeloHostConfError;
  length = epicsSnprintf(request, sizeof(request), "HELO %s\r\n", heloHost);
  if (length >= sizeof(request)) return ImcaEmailStatusSmtpHeloHostConfError;
  return imcaSendSmtpRequestAndReadResponse(client, timeout, request);
}

/* Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpMailRequest(asynUser *client, double timeout,
    const char *from) {
  char request[IMCA_SMTP_REQUEST_SIZE + 1];
  int length;

  if (from == NULL) return ImcaEmailStatusSmtpMailFromError;
  length = epicsSnprintf(request, sizeof(request), "MAIL FROM:<%s>\r\n",
    from);
  if (length >= sizeof(request)) return ImcaEmailStatusSmtpMailFromError;
  return imcaSendSmtpRequestAndReadResponse(client, timeout, request);
}

/* If at least one of the recipient addresses is accepted by the server, this
 * operation is considered a success and returns ImcaEmailStatusSmtpSuccess.
 * Otherwise returns ImcaEmailStatusSmtpRcptToError.
 */
ImcaEmailStatus imcaSendSmtpRcptRequests(asynUser *client, double timeout,
    char *recipients[], int recipientsSize) {
  int atLeastOneOk;
  int i;
  ImcaEmailStatus status;

  atLeastOneOk = 0;
  for (i = 0; i < recipientsSize; i++) {
    if (recipients[i] == NULL) continue;
    if (strcmp(recipients[i], "") == 0) continue;
    status = imcaSendSmtpRcptRequest(client, timeout, recipients[i]);
    if (status == ImcaEmailStatusSmtpSuccess) atLeastOneOk = 1;
  }
  return atLeastOneOk ?
    ImcaEmailStatusSmtpSuccess :
    ImcaEmailStatusSmtpRcptToError;
}

/* recipient must be null-terminated string.
 * Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpRcptRequest(asynUser *client, double timeout,
    const char *recipient) {
  char request[IMCA_SMTP_REQUEST_SIZE + 1];
  int length;

  length = epicsSnprintf(request, sizeof(request), "RCPT TO:<%s>\r\n",
    recipient);
  if (length >= sizeof(request)) return ImcaEmailStatusSmtpRcptToError;
  return imcaSendSmtpRequestAndReadResponse(client, timeout, request);
}

/* Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpDataRequest(asynUser *client, double timeout) {
  return imcaSendSmtpRequestAndReadResponse(client, timeout, "DATA\r\n");
}

/* message must be a properly formatted null-terminated string. I will encode
 * message as an SMTP encoded message and send it to the SMTP server. message
 * must use \012 for newline and must not contain the byte sequence \015\012.
 * message should not contain lines longer than 997 bytes (including the \012
 * newline character) (996 if the line starts with a dot), but I will send
 * messages with lines longer than this even though it violates RFC 821. It is
 * the caller's responsibility to provide a properly formatted message.
 * Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpEncodedMessage(asynUser *client, double timeout,
    const char *message) {
  asynStatus status;
  char buffer[IMCA_SMTP_ENCODED_MESSAGE_SEND_BUFFER_SIZE];
  int bufferLength;
  char current;
  char previous;
  int messageIndex;

  status = pasynOctetSyncIO->flush(client);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;

  buffer[0] = '\0';
  bufferLength = 0;
  previous = '\n';

  for (messageIndex = 0; message[messageIndex] != '\0'; messageIndex++) {
    current = message[messageIndex];
    if (current == '\n') {
      status = imcaAsynBufferedWrite(client, buffer, sizeof(buffer),
        &bufferLength, '\r', timeout);
      if (status != asynSuccess) return ImcaEmailStatusNetworkError;
    } else if (current == '.' && previous == '\n') {
      status = imcaAsynBufferedWrite(client, buffer, sizeof(buffer),
        &bufferLength, '.', timeout);
      if (status != asynSuccess) return ImcaEmailStatusNetworkError;
    }
    status = imcaAsynBufferedWrite(client, buffer, sizeof(buffer),
      &bufferLength, current, timeout);
    if (status != asynSuccess) return ImcaEmailStatusNetworkError;
    previous = current;
  }

  /* partial lines are not allowed */
  if (previous != '\n') return ImcaEmailStatusNetworkError;

  status = imcaAsynWrite(client, buffer, bufferLength, timeout);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;

  status = imcaAsynWriteString(client, ".\r\n", timeout);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;

  return imcaReadSmtpResponse(client, timeout);
}

/* Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpQuitRequest(asynUser *client, double timeout) {
  asynStatus status;

  status = pasynOctetSyncIO->flush(client);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;
  status = imcaAsynWriteString(client, "QUIT\r\n", timeout);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;
  return ImcaEmailStatusSmtpSuccess;
}

/* Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaReadSmtpResponse(asynUser *client, double timeout) {
  asynStatus status;
  char response[IMCA_SMTP_RESPONSE_SIZE + 1];

  for (;;) {
    status = imcaAsynReadString(client, response, sizeof(response), timeout);
    if (status != asynSuccess) return ImcaEmailStatusNetworkError;
    if (imcaIsLastLineOfSmtpResponse(response)) break;
  }
  return imcaGetStatusFromSmtpResponse(response);
}

/* request must include proper "\r\n" termination and be null-terminated.
 * Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaSendSmtpRequestAndReadResponse(asynUser *client,
    double timeout, const char *request) {
  asynStatus status;

  status = pasynOctetSyncIO->flush(client);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;
  status = imcaAsynWriteString(client, request, timeout);
  if (status != asynSuccess) return ImcaEmailStatusNetworkError;
  return imcaReadSmtpResponse(client, timeout);
}

/* response must be a null-terminated string.
 * Returns true if response is last line of SMTP response, false otherwise.
 */
int imcaIsLastLineOfSmtpResponse(const char *response) {
  int i;

  for (i = 0; i < 3; i++) if (response[i] == '\0') return 0;
  return response[3] == ' ' || response[3] == '\0' ? 1 : 0;
}

/* Only looks at first digit of code per recommendation at
 *   http://cr.yp.to/smtp/request.html#response
 * response must be a null-terminated string.
 * Returns ImcaEmailStatusSmtpSuccess on sucess.
 */
ImcaEmailStatus imcaGetStatusFromSmtpResponse(const char *response) {
  char buffer[2];
  long codeFirstDigit;

  buffer[0] = response[0];
  buffer[1] = '\0';

  codeFirstDigit = strtol(buffer, NULL, 10);
  if (codeFirstDigit == 2 || codeFirstDigit == 3)
    return ImcaEmailStatusSmtpSuccess;
  if (codeFirstDigit == 4) return ImcaEmailStatusSmtpDeferral;
  if (codeFirstDigit == 5) return ImcaEmailStatusSmtpFailure;
  return ImcaEmailStatusSmtpSyntaxError;
}

char *imcaGetSmtpHeloHost(void) {
  return getenv(IMCA_SMTP_HELO_HOST_KEY);
}

/* dst must be a null-terminated string.
 * Returns -1 if dst is not big enough, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailDateField(char *dst, size_t dstSize,
    struct tm *dateAsTm) {
  int length;

  length = strftime(dst, dstSize, "Date: %d %b %Y %H:%M:%S %z\n", dateAsTm);
  return length == 0 ? -1 : length;
}

/* dst must be a null-terminated string
 * Returns -1 if dst is not big enough,
 *         -2 if from is empty, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailFromField(char *dst, size_t dstSize, const char *from) {
  int length;

  if (strcmp(from, "") == 0) return -2;
  length = epicsSnprintf(dst, dstSize, "From: %s\n", from);
  return length >= dstSize ? -1 : length;
}

/* dst must be a null-terminated string
 * to values must be null-terminated strings
 * to must contain at least one address
 * Empty email addresses will be silently skipped
 * Returns -1 if dst is not big enough,
 *         -2 if to is empty, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailToField(char *dst, size_t dstSize, char *to[], int toSize) {
  int length;

  length = imcaPrintEmailDestinationAddressField(dst, dstSize, "To", to,
    toSize);
  if (length == -1) return -1;
  if (length == 0) return -2;
  return length;
}

/* dst must be a null-terminated string
 * cc values must be null-terminated strings
 * Empty email addresses will be silently skipped
 * If cc contains no addresses, no field will be written
 * Returns -1 if dst is not big enough, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailCcField(char *dst, size_t dstSize, char *cc[], int ccSize) {
  return imcaPrintEmailDestinationAddressField(dst, dstSize, "Cc", cc,
    ccSize);
}

/* dst must be a null-terminated string
 * subject must be a null-terminated string
 * If subject is empty, no field will be written
 * Returns -1 if dst is not big enough, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailSubjectField(char *dst, size_t dstSize,
    const char *subject) {
  int length;

  if (strcmp(subject, "") == 0) return 0;
  length = epicsSnprintf(dst, dstSize, "Subject: %s\n", subject);
  return length >= dstSize ? -1 : length;
}

/* dst must be a null-terminated string
 * body must be a null-terminated string
 * If body does not end with an empty line, an empty line will be added
 * Returns -1 if dst is not big enough, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailBody(char *dst, size_t dstSize, const char *body) {
  int length;

  if (strcmp(body, "") == 0) return 0;
  length = epicsSnprintf(dst, dstSize, "\n%s%s", body,
    imcaStringEndsWith(body, "\n") ? "" : "\n");
  return length >= dstSize ? -1 : length;
}

/* dst must be a null-terminated string
 * addresses values must be null-terminated strings
 * Empty email addresses will be silently skipped
 * If addresses contains no addresses, no field will be written
 * fieldName must be either "To" or "Cc"
 * Returns -1 if dst is not big enough, otherwise
 *         number of characters printed to dst
 */
int imcaPrintEmailDestinationAddressField(char *dst, size_t dstSize,
    const char *fieldName, char *addresses[], int addressesSize) {
  int isFirst;
  int currentLength;
  size_t currentSize;
  int length;
  int i;

  isFirst = 1;
  currentLength = 0;
  currentSize = dstSize;

  if (addresses == NULL) return currentLength;

  for (i = 0; i < addressesSize; i++) {
    if (strcmp(addresses[i], "") == 0) continue;
    if (isFirst) {
      isFirst = 0;
      length = epicsSnprintf(&dst[currentLength], currentSize, "%.2s: %s",
        fieldName, addresses[i]);
    } else {
      length = epicsSnprintf(&dst[currentLength], currentSize, ",\n    %s",
        addresses[i]);
    }
    if (length >= currentSize) return -1;
    currentLength += length;
    currentSize -= length;
  }

  if (!isFirst) {
    length = epicsSnprintf(&dst[currentLength], currentSize, "\n");
    if (length >= currentSize) return -1;
    currentLength += length;
    currentSize -= length;
  }

  return currentLength;
}
