#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <epicsAssert.h>
#include <epicsEvent.h>
#include <epicsStdio.h>
#include <epicsString.h>
#include <epicsThread.h>
#include <epicsVersion.h>
#include <envDefs.h>
#include <errlog.h>
#include <iocsh.h>
#include <registryFunction.h>
#include <epicsExport.h>

#include <asynDriver.h>
#include <asynOctetSyncIO.h>

#include "imca-asyn-utils.h"
#include "imca-epics-unit.h"
#include "imca-base64.h"
#include "imca-errlog-utils.h"
#include "imca-smtp-client.h"
#include "imca-string-utils.h"
#include "imca-ca-utils.h"

#if defined(VERSION_INT)
#if EPICS_VERSION_INT >= VERSION_INT(3,15,0,2)
#define HAVE_ERRLOG_REMOVE_LISTENERS
#endif
#endif

typedef struct ImcaErrlogExpectContext {
  int index;
  char *(*expectPatterns)[];
  int hadError;
  epicsEventId doneEvent;
} ImcaErrlogExpectContext;

typedef struct ImcaSmtpExpectAndRespondContext {
  char *(*expectationsAndResponses)[];
  asynUser *server;
  double timeout;
  int hadError;
  epicsEventId doneEvent;
} ImcaSmtpExpectAndRespondContext;

static EpicsUnitContext euc;
static char smtpServerAsynPortName[128];
static char smtpClientAsynPortName[128];
static asynUser *smtpClient;
static asynUser *smtpServerConnectionHandler;
static asynUser *smtpServer;
static ImcaErrlogExpectContext errlogExpectContext;
static ImcaSmtpExpectAndRespondContext smtpExpectAndRespondContext;

static void initializeErrlogExpectContext(char *(*expectPatterns)[],
  int isFirstTime);
static asynStatus initializeTestSmtpClientAndServer();
static void errlogExpectListener(void *pvt, const char *message);
static asynStatus destroyTestSmtpClientAndServer();
static void startSmtpExpectAndRespondThread(
  char *(*expectationsAndResponses)[]);
static void smtpExpectAndRespond(ImcaSmtpExpectAndRespondContext *context);
static void initializeSmtpExpectAndRespondContext(int isFirstTime);

static void testImcaEncodeAsBase6401() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "admin:admin";
  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "YWRtaW46YWRtaW4=");
}
static void testImcaEncodeAsBase6402() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "Aladdin:open sesame";
  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "QWxhZGRpbjpvcGVuIHNlc2FtZQ==");
}
static void testImcaEncodeAsBase6403() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "admin:admi";

  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "YWRtaW46YWRtaQ==");
}
static void testImcaEncodeAsBase6404() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "admin:adm";

  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "YWRtaW46YWRt");
}
static void testImcaEncodeAsBase6405() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "";

  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "");
}
static void testImcaEncodeAsBase6406() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "a";

  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "YQ==");
}
static void testImcaEncodeAsBase6407() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "ab";

  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "YWI=");
}
static void testImcaEncodeAsBase6408() {
  char encodedCookie[128];
  char *unencodedCookie;

  unencodedCookie = "abc";

  imcaEncodeAsBase64(encodedCookie, sizeof(encodedCookie), unencodedCookie,
    strlen(unencodedCookie));
  failIfStringNotEqual(&euc, encodedCookie, "YWJj");
}

static void testImcaStripTrailingWhitespace01() {
  char subject[32];

  subject[0] = '\0';
  imcaStripTrailingWhitespace(subject);
  failIfStringNotEqual(&euc, subject, "");
}
static void testImcaStripTrailingWhitespace02() {
  char subject[32];

  strcpy(subject, "a\n");
  imcaStripTrailingWhitespace(subject);
  failIfStringNotEqual(&euc, subject, "a");
}
static void testImcaStripTrailingWhitespace03() {
  char subject[32];

  strcpy(subject, "a\r\n");
  imcaStripTrailingWhitespace(subject);
  failIfStringNotEqual(&euc, subject, "a");
}
static void testImcaStripTrailingWhitespace04() {
  char subject[32];

  strcpy(subject, "a ");
  imcaStripTrailingWhitespace(subject);
  failIfStringNotEqual(&euc, subject, "a");
}
static void testImcaStripTrailingWhitespace05() {
  char subject[32];

  strcpy(subject, "a b ");
  imcaStripTrailingWhitespace(subject);
  failIfStringNotEqual(&euc, subject, "a b");
}
static void testImcaStripTrailingWhitespace06() {
  char subject[32];

  strcpy(subject, "</html>");
  imcaStripTrailingWhitespace(subject);
  failIfStringNotEqual(&euc, subject, "</html>");
}

static void testImcaStringEndsWith01() {
  failIf(&euc, !imcaStringEndsWith("a", "a"));
}
static void testImcaStringEndsWith02() {
  failIf(&euc, !imcaStringEndsWith("ab", "b"));
}
static void testImcaStringEndsWith03() {
  failIf(&euc, imcaStringEndsWith("a", "aa"));
}
static void testImcaStringEndsWith04() {
  failIf(&euc, !imcaStringEndsWith("abc", "bc"));
}
static void testImcaStringEndsWith05() {
  failIf(&euc, imcaStringEndsWith("abc", "ab"));
}
static void testImcaStringEndsWith06() {
  failIf(&euc, !imcaStringEndsWith("abcdefg", "efg"));
}

static void testImcaStringEndsWithIgnoringCase01() {
  failIf(&euc, !imcaStringEndsWithIgnoringCase("</BODY></HTML>", "</html>"));
}
static void testImcaStringEndsWithIgnoringCase02() {
  char response[1024];

  sprintf(response,
    "HTTP/1.0 302 Found\r\n"
    "Location: /index.shtml\r\n"
    "Date: Wed, 27 Sep 2006 6:41:38 GMT\r\n"
    "Connection: close\r\n"
    "Content-Type: text/html\r\n"
    "\r\n"
    "<HTML>  <HEAD> <TITLE>document has moved</TITLE></HEAD>\r\n"
    "<BODY>document has moved to <A HREF=\"/index.shtml\">here</A></BODY>\r\n"
    "</HTML>");
  failIf(&euc, !imcaStringEndsWithIgnoringCase(response, "</html>"));
}

static void testImcaWaitForIpServerConnection01() {
  /* returns epicsEventWaitTimeout after waiting timeout for connection */
  asynStatus status;
  epicsEventWaitStatus waitStatus;

  smtpServer = NULL;
  smtpServerConnectionHandler = NULL;

  status = imcaCreateListeningIpServer(&smtpServerConnectionHandler,
    smtpServerAsynPortName, "\n", 1, "", 0);
  failIfIntNotEqual(&euc, asynSuccess, status);
  if (status != asynSuccess) return;

  waitStatus = imcaWaitForIpServerConnection(&smtpServer,
    smtpServerConnectionHandler, 2.0);
  failIfIntNotEqual(&euc, epicsEventWaitTimeout, waitStatus);
  if (waitStatus == epicsEventWaitOK) imcaDestroyIpServerSession(smtpServer);

  status = imcaDestroyIpServer(smtpServerConnectionHandler);
  failIfIntNotEqual(&euc, asynSuccess, status);
}

static void testImcaPrintEmailDateField01() {
  char expected[80];
  struct tm dateAsTm;
  time_t dateAsSeconds;
  char dateField[80];
  int length;

  dateAsSeconds = 1167970245;
  gmtime_r(&dateAsSeconds, &dateAsTm);
  strftime(expected, sizeof(expected), "Date: %d %b %Y %H:%M:%S %z\n",
    &dateAsTm);
  dateField[0] = '\0';
  length = imcaPrintEmailDateField(dateField, sizeof(dateField), &dateAsTm);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, dateField, expected);
}
static void testImcaPrintEmailDateField02() {
  char expected[80];
  struct tm dateAsTm;
  time_t dateAsSeconds;
  char dateField[80];
  int length;

  dateAsSeconds = 1167970245;
  gmtime_r(&dateAsSeconds, &dateAsTm);
  strftime(expected, sizeof(expected), "Date: %d %b %Y %H:%M:%S %z\n",
    &dateAsTm);
  dateField[0] = '\0';
  length = imcaPrintEmailDateField(dateField, strlen(expected) - 1,
    &dateAsTm);
  failIfIntNotEqual(&euc, length, -1);
}

static void testImcaPrintEmailFromField01() {
  char *from;
  char expected[80];
  char fromField[80];
  int length;

  from = "\"17-BM Instrument Server\" <admin@imca.aps.anl.gov>";
  sprintf(expected, "From: %s\n", from);
  fromField[0] = '\0';
  length = imcaPrintEmailFromField(fromField, sizeof(fromField), from);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, fromField, expected);
}
static void testImcaPrintEmailFromField02() {
  /* buffer too short */
  char *from;
  char expected[80];
  char fromField[80];
  int length;

  from = "\"17-BM Instrument Server\" <admin@imca.aps.anl.gov>";
  sprintf(expected, "From: %s\n", from);
  fromField[0] = '\0';
  length = imcaPrintEmailFromField(fromField, strlen(expected) - 1, from);
  failIfIntNotEqual(&euc, length, -1);
}
static void testImcaPrintEmailFromField03() {
  /* empty from */
  char *from;
  char expected[80];
  char fromField[80];
  int length;

  from = "";
  sprintf(expected, "From: %s\n", from);
  fromField[0] = '\0';
  length = imcaPrintEmailFromField(fromField, sizeof(fromField), from);
  failIfIntNotEqual(&euc, length, -2);
}

static void testImcaPrintEmailToField01() {
  char *to[1];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";

  sprintf(expected, "To: %s\n", to[0]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, toField, expected);
}
static void testImcaPrintEmailToField02() {
  /* 2 address */
  char *to[2];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";
  to[1] = "yoda@imca.aps.anl.gov";

  sprintf(expected, "To: %s,\n"
                    "    %s\n", to[0], to[1]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, toField, expected);
}
static void testImcaPrintEmailToField03() {
  /* 3 address */
  char *to[3];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";
  to[1] = "yoda@imca.aps.anl.gov";
  to[2] = "han@imca.aps.anl.gov";

  sprintf(expected, "To: %s,\n"
                    "    %s,\n"
                    "    %s\n", to[0], to[1], to[2]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, toField, expected);
}
static void testImcaPrintEmailToField04() {
  /* Destination buffer is not big enough */
  char *to[1];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";

  sprintf(expected, "To: %s\n", to[0]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, strlen(expected) - 1, to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, -1);
}
static void testImcaPrintEmailToField05() {
  /* Destination buffer is not big enough for 2 addresses */
  char *to[2];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";
  to[1] = "yoda@imca.aps.anl.gov";

  sprintf(expected, "To: %s,\n"
                    "    %s\n", to[0], to[1]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, strlen(to[0]) + strlen("To: ") + 1,
    to, sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, -1);
}
static void testImcaPrintEmailToField06() {
  /* zero length To list so empty To */
  char *to[0];
  char toField[80];
  int length;

  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to, 0);
  failIfIntNotEqual(&euc, length, -2);
}
static void testImcaPrintEmailToField07() {
  /* non-zero length To list, but no addresses so empty To */
  char *to[1];
  char toField[80];
  int length;

  to[0] = "";

  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, -2);
}
static void testImcaPrintEmailToField08() {
  /* To list with some empty, should silently skip the empty addresses */
  char *to[3];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";
  to[1] = "";
  to[2] = "yoda@imca.aps.anl.gov";

  sprintf(expected, "To: %s,\n"
                    "    %s\n", to[0], to[2]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, toField, expected);
}
static void testImcaPrintEmailToField09() {
  /* To list with later empty, should silently skip the empty addresses */
  char *to[3];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";
  to[1] = "";
  to[2] = "";

  sprintf(expected, "To: %s\n", to[0]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, toField, expected);
}
static void testImcaPrintEmailToField10() {
  /* Just enough room for one address */
  char *to[3];
  char expected[80];
  char toField[80];
  int length;

  to[0] = "luke@imca.aps.anl.gov";
  to[1] = "";
  to[2] = "";

  sprintf(expected, "To: %s\n", to[0]);
  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, strlen(expected) + 1, to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, toField, expected);
}
static void testImcaPrintEmailToField11() {
  /* empty dst */
  char *to[1];
  char toField[1];
  int length;

  to[0] = "luke@imca.aps.anl.gov";

  toField[0] = '\0';
  length = imcaPrintEmailToField(toField, sizeof(toField), to,
    sizeof(to) / sizeof(to[0]));
  failIfIntNotEqual(&euc, length, -1);
}

static void testImcaPrintEmailCcField01() {
  /* 1 address */
  char *cc[1];
  char expected[80];
  char ccField[80];
  int length;

  cc[0] = "luke@imca.aps.anl.gov";

  sprintf(expected, "Cc: %s\n", cc[0]);
  ccField[0] = '\0';
  length = imcaPrintEmailCcField(ccField, sizeof(ccField), cc,
    sizeof(cc) / sizeof(cc[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, ccField, expected);
}
static void testImcaPrintEmailCcField02() {
  /* Cc field will be left out if there are no addresses for it */
  char *cc[1];
  char expected[80];
  char ccField[80];
  int length;

  cc[0] = "";

  expected[0] = '\0';
  ccField[0] = '\0';
  length = imcaPrintEmailCcField(ccField, sizeof(ccField), cc,
    sizeof(cc) / sizeof(cc[0]));
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, ccField, expected);
}
static void testImcaPrintEmailCcField03() {
  /* NULL cc value. Field will be left out. */
  char expected[80];
  char ccField[80];
  int length;

  expected[0] = '\0';
  ccField[0] = '\0';
  length = imcaPrintEmailCcField(ccField, sizeof(ccField), NULL, 0);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, ccField, expected);
}

static void testImcaPrintEmailSubjectField01() {
  char *subject;
  char expected[80];
  char subjectField[80];
  int length;

  subject = "Data collection paused";
  sprintf(expected, "Subject: %s\n", subject);
  subjectField[0] = '\0';
  length = imcaPrintEmailSubjectField(subjectField, sizeof(subjectField),
    subject);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, subjectField, expected);
}
static void testImcaPrintEmailSubjectField02() {
  /* buffer too short */
  char *subject;
  char expected[80];
  char subjectField[80];
  int length;

  subject = "Data collection paused";
  sprintf(expected, "Subject: %s\n", subject);
  subjectField[0] = '\0';
  length = imcaPrintEmailSubjectField(subjectField, strlen(expected),
    subject);
  failIfIntNotEqual(&euc, length, -1);
}
static void testImcaPrintEmailSubjectField03() {
  /* empty subject */
  char *subject;
  char expected[80];
  char subjectField[80];
  int length;

  subject = "";
  expected[0] = '\0';
  subjectField[0] = '\0';
  length = imcaPrintEmailSubjectField(subjectField, sizeof(subjectField),
    subject);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, subjectField, expected);
}

static void testImcaPrintEmailBody01() {
  char *body;
  char expected[80];
  char bodyPart[80];
  int length;

  body = "* Robot dewar #1 level is low\n";
  sprintf(expected, "\n%s", body);
  bodyPart[0] = '\0';
  length = imcaPrintEmailBody(bodyPart, sizeof(bodyPart), body);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, bodyPart, expected);
}
static void testImcaPrintEmailBody02() {
  /* buffer too short */
  char *body;
  char expected[80];
  char bodyPart[80];
  int length;

  body = "* Robot dewar #1 level is low\n";
  sprintf(expected, "\n%s", body);
  bodyPart[0] = '\0';
  length = imcaPrintEmailBody(bodyPart, strlen(expected), body);
  failIfIntNotEqual(&euc, length, -1);
}
static void testImcaPrintEmailBody03() {
  /* empty body */
  char *body;
  char expected[80];
  char bodyPart[80];
  int length;

  body = "";
  expected[0] = '\0';
  bodyPart[0] = '\0';
  length = imcaPrintEmailBody(bodyPart, sizeof(bodyPart), body);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, bodyPart, expected);
}
static void testImcaPrintEmailBody04() {
  /* multi-line body */
  char *body;
  char expected[80];
  char bodyPart[80];
  int length;

  body = "* Ring is down\n"
         "* Robot dewar #1 level is low\n"
         "\n"
         "17-BM Instrument Server\n";
  sprintf(expected, "\n%s", body);
  bodyPart[0] = '\0';
  length = imcaPrintEmailBody(bodyPart, sizeof(bodyPart), body);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, bodyPart, expected);
}
static void testImcaPrintEmailBody05() {
  /* body that has a non-empty last line. an empty line will be added. */
  char *body;
  char expected[80];
  char bodyPart[80];
  int length;

  body = "* Robot dewar #1 level is low";
  sprintf(expected, "\n%s\n", body);
  bodyPart[0] = '\0';
  length = imcaPrintEmailBody(bodyPart, sizeof(bodyPart), body);
  failIfIntNotEqual(&euc, length, strlen(expected));
  failIfStringNotEqual(&euc, bodyPart, expected);
}

static void testImcaCreateEmailMessage01() {
  char *dateAsString;
  char *from;
  char *to[1];
  char *subject;
  char *body;
  char expected[512];
  char message[512];
  ImcaEmailStatus status;
  char dateField[80];

  dateAsString = "05 Jan 2007 04:10:45 +0000";
  from = "\"17-BM Instrument Server\" <admin@imca.aps.anl.gov>";
  to[0] = "luke@imca.aps.anl.gov";
  subject = "Data collection paused";
  body = "* Robot dewar #1 level is low\n"
         "\n"
         "17-BM Instrument Server\n";

  expected[0] = '\0';
  sprintf(expected, "Date: %s\n"
                    "From: %s\n"
                    "To: %s\n"
                    "Subject: %s\n"
                    "\n"
                    "%s", dateAsString, from, to[0], subject, body);

  message[0] = '\0';
  status = imcaCreateEmailMessage(message, sizeof(message), from, to,
    sizeof(to) / sizeof(to[0]), NULL, 0, subject, body);
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSuccess);

  /* overwrite generated date to easily test equality */
  dateField[0] = '\0';
  sprintf(dateField, "Date: %s\n", dateAsString);
  bcopy(dateField, message, strlen(dateField));

  failIfStringNotEqual(&euc, message, expected);
}
static void testImcaCreateEmailMessage02() {
  /* with Cc field */
  char *dateAsString;
  char *from;
  char *to[1];
  char *cc[1];
  char *subject;
  char *body;
  char expected[512];
  char message[512];
  ImcaEmailStatus status;
  char dateField[80];

  dateAsString = "05 Jan 2007 04:10:45 +0000";
  from = "\"17-BM Instrument Server\" <admin@imca.aps.anl.gov>";
  to[0] = "luke@imca.aps.anl.gov";
  cc[0] = "yoda@imca.aps.anl.gov";
  subject = "Data collection paused";
  body = "* Robot dewar #1 level is low\n"
         "\n"
         "17-BM Instrument Server\n";

  expected[0] = '\0';
  sprintf(expected, "Date: %s\n"
                    "From: %s\n"
                    "To: %s\n"
                    "Cc: %s\n"
                    "Subject: %s\n"
                    "\n"
                    "%s", dateAsString, from, to[0], cc[0], subject, body);

  message[0] = '\0';
  status = imcaCreateEmailMessage(message, sizeof(message), from, to,
    sizeof(to) / sizeof(to[0]), cc, sizeof(cc) / sizeof(cc[0]), subject,
    body);
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSuccess);

  /* overwrite generated date to easily test equality */
  dateField[0] = '\0';
  sprintf(dateField, "Date: %s\n", dateAsString);
  bcopy(dateField, message, strlen(dateField));

  failIfStringNotEqual(&euc, message, expected);
}
static void testImcaCreateEmailMessage03() {
  /* Empty body */
  char *dateAsString;
  char *from;
  char *to[1];
  char *subject;
  char *body;
  char expected[512];
  char message[512];
  ImcaEmailStatus status;
  char dateField[80];

  dateAsString = "05 Jan 2007 04:10:45 +0000";
  from = "\"17-BM Instrument Server\" <admin@imca.aps.anl.gov>";
  to[0] = "luke@imca.aps.anl.gov";
  subject = "Data collection paused";
  body = "";

  expected[0] = '\0';
  sprintf(expected, "Date: %s\n"
                    "From: %s\n"
                    "To: %s\n"
                    "Subject: %s\n", dateAsString, from, to[0], subject);

  message[0] = '\0';
  status = imcaCreateEmailMessage(message, sizeof(message), from, to,
    sizeof(to) / sizeof(to[0]), NULL, 0, subject, body);
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSuccess);

  /* overwrite generated date to easily test equality */
  dateField[0] = '\0';
  sprintf(dateField, "Date: %s\n", dateAsString);
  bcopy(dateField, message, strlen(dateField));

  failIfStringNotEqual(&euc, message, expected);
}

static void testImcaGetStatusFromSmtpResponse01() {
  ImcaEmailStatus status;

  status = imcaGetStatusFromSmtpResponse("220 ok");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpSuccess);
  status = imcaGetStatusFromSmtpResponse("354 ok");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpSuccess);
  status = imcaGetStatusFromSmtpResponse("450 deferral");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpDeferral);
  status = imcaGetStatusFromSmtpResponse("500 failure");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpFailure);
  status = imcaGetStatusFromSmtpResponse("600 syntax error");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpSyntaxError);
  /* no space after 3 digit code even though there should be */
  status = imcaGetStatusFromSmtpResponse("250");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpSuccess);
  /* response line that is not last line */
  status = imcaGetStatusFromSmtpResponse("250-PIPELINING");
  failIfIntNotEqual(&euc, status, ImcaEmailStatusSmtpSuccess);
}

static void testImcaIsLastLineOfSmtpResponse01() {
  failIf(&euc, imcaIsLastLineOfSmtpResponse("250-PIPELINING"));
  failIf(&euc, !imcaIsLastLineOfSmtpResponse("250 PIPELINING"));
  /* some SMTP servers incorrectly leave off the space */
  failIf(&euc, !imcaIsLastLineOfSmtpResponse("250"));
}

static void testImcaReadSmtpResponse01() {
  /* single line response */
  char *response;
  asynStatus astatus;
  ImcaEmailStatus estatus;

  response = "220 localhost\r\n";

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  astatus = imcaAsynWriteString(smtpServer, response, 0.5);
  failIfIntNotEqual(&euc, astatus, asynSuccess);
  estatus = imcaReadSmtpResponse(smtpClient, 0.5);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  estatus = imcaReadSmtpResponse(smtpClient, 0.5);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusNetworkError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaReadSmtpResponse02() {
  /* multi-line response */
  char *response;
  asynStatus astatus;
  ImcaEmailStatus estatus;

  response = "250-localhost\r\n"
             "250-PIPELINING\r\n"
             "250 8BITMIME\r\n";

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  astatus = imcaAsynWriteString(smtpServer, response, 1.0);
  failIfIntNotEqual(&euc, astatus, asynSuccess);
  estatus = imcaReadSmtpResponse(smtpClient, 0.5);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  estatus = imcaReadSmtpResponse(smtpClient, 0.5);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusNetworkError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaWaitForSmtpServerGreeting01() {
  asynStatus astatus;
  ImcaEmailStatus estatus;

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  astatus = imcaAsynWriteString(smtpServer, "220 localhost\r\n", 1.0);
  failIfIntNotEqual(&euc, astatus, asynSuccess);
  estatus = imcaWaitForSmtpServerGreeting(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaSendSmtpHeloRequest01() {
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "HELO [127.0.0.1]\r", "250 localhost\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpHeloRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpHeloRequest02() {
  /* server rejects HELO request with a deferral */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "HELO [127.0.0.1]\r", "421 deferral\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpHeloRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpDeferral);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpHeloRequest03() {
  /* server rejects HELO request with a failure */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "HELO [127.0.0.1]\r", "500 failure\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpHeloRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpFailure);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpHeloRequest04() {
  /* HELO host env var is too long */
  asynStatus astatus;
  char *origHeloHost;
  int i;
  char longHeloHost[IMCA_SMTP_REQUEST_SIZE + 1];
  ImcaEmailStatus estatus;

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  origHeloHost = getenv(IMCA_SMTP_HELO_HOST_KEY);
  for (i = 0; i < sizeof(longHeloHost); i++) longHeloHost[i] = 'x';
  longHeloHost[sizeof(longHeloHost) - 1] = '\0';
  /* NOTE: epicsEnvSet leaks memory by design */
  epicsEnvSet(IMCA_SMTP_HELO_HOST_KEY, longHeloHost);
  estatus = imcaSendSmtpHeloRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpHeloHostConfError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
  /* NOTE: epicsEnvSet leaks memory by design */
  epicsEnvSet(IMCA_SMTP_HELO_HOST_KEY, origHeloHost);
}

static void testImcaSendSmtpMailRequest01() {
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *from = "sysadmin@imca.aps.anl.gov";
  char *expect[] = {
    "MAIL FROM:<sysadmin@imca.aps.anl.gov>\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpMailRequest(smtpClient, 1.0, from);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpMailRequest02() {
  /* deferral */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *from = "sysadmin@imca.aps.anl.gov";
  char *expect[] = {
    "MAIL FROM:<sysadmin@imca.aps.anl.gov>\r", "451 timeout\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpMailRequest(smtpClient, 1.0, from);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpDeferral);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpMailRequest03() {
  /* failure */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *from = "sysadmin@imca.aps.anl.gov";
  char *expect[] = {
    "MAIL FROM:<sysadmin@imca.aps.anl.gov>\r", "500 line too long\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpMailRequest(smtpClient, 1.0, from);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpFailure);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpMailRequest04() {
  /* from is too long */
  asynStatus astatus;
  int i;
  char from[IMCA_SMTP_REQUEST_SIZE + 1];
  ImcaEmailStatus estatus;

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  for (i = 0; i < sizeof(from); i++) from[i] = 'x';
  from[sizeof(from) - 1] = '\0';
  estatus = imcaSendSmtpMailRequest(smtpClient, 1.0, from);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpMailFromError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaSendSmtpRcptRequests01() {
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {
    "yoda@imca.aps.anl.gov"
  };
  int recipientsSize = 1;
  char *expect[] = {
    "RCPT TO:<yoda@imca.aps.anl.gov>\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests02() {
  /* More than 1 recipient */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {
    "yoda@imca.aps.anl.gov",
    "luke@imca.aps.anl.gov"
  };
  int recipientsSize = 2;
  char *expect[] = {
    "RCPT TO:<yoda@imca.aps.anl.gov>\r", "250 ok\r\n",
    "RCPT TO:<luke@imca.aps.anl.gov>\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests03() {
  /* Some good but some Empty and NULL recipients */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {
    "yoda@imca.aps.anl.gov",
    NULL,
    "luke@imca.aps.anl.gov",
    "",
    "han@imca.aps.anl.gov"
  };
  int recipientsSize = 5;
  char *expect[] = {
    "RCPT TO:<yoda@imca.aps.anl.gov>\r", "250 ok\r\n",
    "RCPT TO:<luke@imca.aps.anl.gov>\r", "250 ok\r\n",
    "RCPT TO:<han@imca.aps.anl.gov>\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests04() {
  /* Some accepted and some rejected */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {
    "yoda@imca.aps.anl.gov",
    "luke@imca.aps.anl.gov",
    "han@imca.aps.anl.gov"
  };
  int recipientsSize = 3;
  char *expect[] = {
    "RCPT TO:<yoda@imca.aps.anl.gov>\r", "550 no such recipient here\r\n",
    "RCPT TO:<luke@imca.aps.anl.gov>\r", "250 ok\r\n",
    "RCPT TO:<han@imca.aps.anl.gov>\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests05() {
  /* All rejected */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {
    "yoda@imca.aps.anl.gov",
    "luke@imca.aps.anl.gov"
  };
  int recipientsSize = 2;
  char *expect[] = {
    "RCPT TO:<yoda@imca.aps.anl.gov>\r", "550 no such recipient here\r\n",
    "RCPT TO:<luke@imca.aps.anl.gov>\r", "550 no such recipient here\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpRcptToError);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests06() {
  /* All empty */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {
    "",
    NULL
  };
  int recipientsSize = 2;

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpRcptToError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests07() {
  /* Empty recipients list */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *recipients[] = {};
  int recipientsSize = 0;

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpRcptToError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpRcptRequests08() {
  /* One too long but other is fine */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  int i;
  char longRecipient[IMCA_SMTP_REQUEST_SIZE + 1];
  char *recipients[2];
  int recipientsSize = 2;
  char *expect[] = {
    "RCPT TO:<yoda@imca.aps.anl.gov>\r", "250 ok\r\n",
    NULL
  };

  for (i = 0; i < sizeof(longRecipient); i++) longRecipient[i] = 'x';
  longRecipient[sizeof(longRecipient) - 1] = '\0';
  recipients[0] = longRecipient;
  recipients[1] = "yoda@imca.aps.anl.gov";

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpRcptRequests(smtpClient, 1.0, recipients,
    recipientsSize);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaSendSmtpDataRequest01() {
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "DATA\r", "354 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpDataRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpDataRequest02() {
  /* deferral */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "DATA\r", "451 unable to set up internal message buffer\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpDataRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpDeferral);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpDataRequest03() {
  /* failure */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "DATA\r", "503 need RCPT before DATA\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpDataRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpFailure);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaSendSmtpEncodedMessage01() {
  /* basic */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message =
    "Date: 05 Jan 2007 04:10:45 +0000\n"
    "From: yoda@imca.aps.anl.gov\n"
    "To: luke@imca.aps.anl.gov\n"
    "Subject: Well wishes\n"
    "\n"
    "May the force be with you.\n";
  char *expect[] = {
    "Date: 05 Jan 2007 04:10:45 +0000\r", "",
    "From: yoda@imca.aps.anl.gov\r", "",
    "To: luke@imca.aps.anl.gov\r", "",
    "Subject: Well wishes\r", "",
    "\r", "",
    "May the force be with you.\r", "",
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage02() {
  /* more than one line of body */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message =
    "Date: 05 Jan 2007 04:10:45 +0000\n"
    "From: yoda@imca.aps.anl.gov\n"
    "To: luke@imca.aps.anl.gov\n"
    "Subject: Well wishes\n"
    "\n"
    "May the force be with you.\n"
    "-Yoda\n";
  char *expect[] = {
    "Date: 05 Jan 2007 04:10:45 +0000\r", "",
    "From: yoda@imca.aps.anl.gov\r", "",
    "To: luke@imca.aps.anl.gov\r", "",
    "Subject: Well wishes\r", "",
    "\r", "",
    "May the force be with you.\r", "",
    "-Yoda\r", "",
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage03() {
  /* zero-line message */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message = "";
  char *expect[] = {
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage04() {
  /* empty line */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message = "\n";
  char *expect[] = {
    "\r", "",
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage05() {
  /* line starting with dot */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message = ".\n";
  char *expect[] = {
    "..\r", "",
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage06() {
  /* line starting with dot and line with a non-starting dot */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message =
    "abc\n"
    ".dot\n"
    "another.dot\n";
  char *expect[] = {
    "abc\r", "",
    "..dot\r", "",
    "another.dot\r", "",
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage07() {
  /* consider message I/O error for message that has a partial line */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message = "abc";
  time_t startTime;
  time_t endTime;

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startTime = time(NULL);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 3.0, message);
  endTime = time(NULL);
  failIf(&euc, endTime - startTime > 2.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusNetworkError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage08() {
  /* line longer than send buffer should be sent in two parts */
  asynStatus astatus;
  int i;
  ImcaEmailStatus estatus;
  char message[IMCA_SMTP_ENCODED_MESSAGE_SEND_BUFFER_SIZE + 5];
  char expectedMessage[IMCA_SMTP_ENCODED_MESSAGE_SEND_BUFFER_SIZE + 5];
  char *expect[] = {
    "", "",
    ".\r", "250 ok\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  for (i = 0; i < sizeof(message) - 2; i++) message[i] = 'x';
  message[sizeof(message) - 2] = '\n';
  message[sizeof(message) - 1] = '\0';
  for (i = 0; i < sizeof(expectedMessage) - 2; i++) expectedMessage[i] = 'x';
  expectedMessage[sizeof(expectedMessage) - 2] = '\r';
  expectedMessage[sizeof(expectedMessage) - 1] = '\0';
  expect[0] = expectedMessage;
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpEncodedMessage09() {
  /* server rejects message with a failure code */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *message = "abc\n";
  char *expect[] = {
    "abc\r", "",
    ".\r", "552 sorry, this message is too large for my disk\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpEncodedMessage(smtpClient, 1.0, message);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpFailure);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaSendSmtpQuitRequest01() {
  /* basic */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "QUIT\r", "221 goodbye\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpQuitRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendSmtpQuitRequest02() {
  /* don't wait for server response */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *expect[] = {
    "QUIT\r", "504 I'm trying to tell you you can't quit\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendSmtpQuitRequest(smtpClient, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaSendEmail01() {
  /* basic */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *from = "yoda@imca.aps.anl.gov";
  int recipientsSize = 1;
  char *recipients[] = { "luke@imca.aps.anl.gov" };
  char *message =
    "Date: 05 Jan 2007 04:10:45 +0000\n"
    "From: yoda@imca.aps.anl.gov\n"
    "To: luke@imca.aps.anl.gov\n"
    "Subject: Well wishes\n"
    "\n"
    "May the force be with you.\n";
  char *expect[] = {
    "", "220 localhost\r\n",
    "HELO [127.0.0.1]\r", "250 localhost\r\n",
    "MAIL FROM:<yoda@imca.aps.anl.gov>\r", "250 ok\r\n",
    "RCPT TO:<luke@imca.aps.anl.gov>\r", "250 ok\r\n",
    "DATA\r", "354 go ahead\r\n",
    "Date: 05 Jan 2007 04:10:45 +0000\r", "",
    "From: yoda@imca.aps.anl.gov\r", "",
    "To: luke@imca.aps.anl.gov\r", "",
    "Subject: Well wishes\r", "",
    "\r", "",
    "May the force be with you.\r", "",
    ".\r", "250 ok\r\n",
    "QUIT\r", "221 goodbye\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendEmailUsingAsynUser(smtpClient, from, recipients,
    recipientsSize, message, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSuccess);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendEmail02() {
  /* Timeout waiting for SMTP greeting */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *from = "yoda@imca.aps.anl.gov";
  int recipientsSize = 1;
  char *recipients[] = { "luke@imca.aps.anl.gov" };
  char *message =
    "Date: 05 Jan 2007 04:10:45 +0000\n"
    "From: yoda@imca.aps.anl.gov\n"
    "To: luke@imca.aps.anl.gov\n"
    "Subject: Well wishes\n"
    "\n"
    "May the force be with you.\n";
  char *expect[] = {
    "QUIT\r", "",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendEmailUsingAsynUser(smtpClient, from, recipients,
    recipientsSize, message, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusNetworkError);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}
static void testImcaSendEmail03() {
  /* SMTP RCPT failure */
  asynStatus astatus;
  ImcaEmailStatus estatus;
  char *from = "yoda@imca.aps.anl.gov";
  int recipientsSize = 1;
  char *recipients[] = { "luke@imca.aps.anl.gov" };
  char *message =
    "Date: 05 Jan 2007 04:10:45 +0000\n"
    "From: yoda@imca.aps.anl.gov\n"
    "To: luke@imca.aps.anl.gov\n"
    "Subject: Well wishes\n"
    "\n"
    "May the force be with you.\n";
  char *expect[] = {
    "", "220 localhost\r\n",
    "HELO [127.0.0.1]\r", "250 localhost\r\n",
    "MAIL FROM:<yoda@imca.aps.anl.gov>\r", "250 ok\r\n",
    "RCPT TO:<luke@imca.aps.anl.gov>\r", "550 no such recipient here\r\n",
    "QUIT\r", "221 goodbye\r\n",
    NULL
  };

  astatus = initializeTestSmtpClientAndServer();
  if (astatus != asynSuccess) { failIf(&euc, 1); return; }
  startSmtpExpectAndRespondThread(&expect);
  estatus = imcaSendEmailUsingAsynUser(smtpClient, from, recipients,
    recipientsSize, message, 1.0);
  failIfIntNotEqual(&euc, estatus, ImcaEmailStatusSmtpRcptToError);
  epicsEventWait(smtpExpectAndRespondContext.doneEvent);
  failIf(&euc, smtpExpectAndRespondContext.hadError);
  astatus = destroyTestSmtpClientAndServer();
  failIfIntNotEqual(&euc, astatus, asynSuccess);
}

static void testImcaErrlogTsPrintf01() {
  int i;
  char message[233];
  char messagePattern[235];
  epicsEventWaitStatus status;
  char *expectPatterns[10];

  /* basic */

  expectPatterns[0] = "* ";
  expectPatterns[1] = "* \n";
  expectPatterns[2] = "* a\n";
  expectPatterns[3] = "* a\n";
  expectPatterns[4] = "* a 3\n";
  expectPatterns[5] = NULL;

  initializeErrlogExpectContext(&expectPatterns, 0);
  errlogAddListener(errlogExpectListener, (void *)&errlogExpectContext);
  eltc(0);

  imcaErrlogTsPrintf("");
  imcaErrlogTsPrintf("%s\n", "");
  imcaErrlogTsPrintf("a\n");
  imcaErrlogTsPrintf("%s\n", "a");
  imcaErrlogTsPrintf("%s %d\n", "a", 3);

  status = epicsEventWaitWithTimeout(errlogExpectContext.doneEvent, 5.0);
  #ifdef HAVE_ERRLOG_REMOVE_LISTENERS
  errlogRemoveListeners(errlogExpectListener, (void *)&errlogExpectContext);
  #else
  errlogRemoveListener(errlogExpectListener);
  #endif
  eltc(1);

  failIf(&euc, status != epicsEventWaitOK);
  failIf(&euc, errlogExpectContext.hadError);

  /* largest that should fit in buffer (maxBufferLen - timestampLen -
   * spaceCharacterSize = 255 - 23 - 1 = 231) */

  for (i = 0; i < 231; i++) message[i] = i == 230 ? '\n' : 'x';
  message[231] = '\0';
  strncpy(messagePattern, "* ", sizeof(messagePattern) - 1);
  messagePattern[sizeof(messagePattern) - 1] = '\0';
  strncat(messagePattern, message, sizeof(messagePattern) -
    strlen(messagePattern) - 1);

  expectPatterns[0] = messagePattern;
  expectPatterns[1] = messagePattern;
  expectPatterns[2] = NULL;

  initializeErrlogExpectContext(&expectPatterns, 0);
  errlogAddListener(errlogExpectListener, (void *)&errlogExpectContext);
  eltc(0);

  imcaErrlogTsPrintf(message);
  imcaErrlogTsPrintf("%s", message);

  status = epicsEventWaitWithTimeout(errlogExpectContext.doneEvent, 5.0);
  #ifdef HAVE_ERRLOG_REMOVE_LISTENERS
  errlogRemoveListeners(errlogExpectListener, (void *)&errlogExpectContext);
  #else
  errlogRemoveListener(errlogExpectListener);
  #endif
  eltc(1);

  failIf(&euc, status != epicsEventWaitOK);
  failIf(&euc, errlogExpectContext.hadError);

  /* shouldn't fit in buffer; will be written in two parts: first the
   * timestamp, then the message */

  for (i = 0; i < 232; i++) message[i] = i == 231 ? '\n' : 'x';
  message[232] = '\0';

  expectPatterns[0] = "* ";
  expectPatterns[1] = message;
  expectPatterns[2] = "* ";
  expectPatterns[3] = message;
  expectPatterns[4] = NULL;

  initializeErrlogExpectContext(&expectPatterns, 0);
  errlogAddListener(errlogExpectListener, (void *)&errlogExpectContext);
  eltc(0);

  imcaErrlogTsPrintf(message);
  imcaErrlogTsPrintf("%s", message);

  status = epicsEventWaitWithTimeout(errlogExpectContext.doneEvent, 5.0);
  #ifdef HAVE_ERRLOG_REMOVE_LISTENERS
  errlogRemoveListeners(errlogExpectListener, (void *)&errlogExpectContext);
  #else
  errlogRemoveListener(errlogExpectListener);
  #endif
  eltc(1);

  failIf(&euc, status != epicsEventWaitOK);
  failIf(&euc, errlogExpectContext.hadError);
}

static void runAllTests() {
  runTest(&euc, testImcaEncodeAsBase6401);
  runTest(&euc, testImcaEncodeAsBase6402);
  runTest(&euc, testImcaEncodeAsBase6403);
  runTest(&euc, testImcaEncodeAsBase6404);
  runTest(&euc, testImcaEncodeAsBase6405);
  runTest(&euc, testImcaEncodeAsBase6406);
  runTest(&euc, testImcaEncodeAsBase6407);
  runTest(&euc, testImcaEncodeAsBase6408);

  runTest(&euc, testImcaStripTrailingWhitespace01);
  runTest(&euc, testImcaStripTrailingWhitespace02);
  runTest(&euc, testImcaStripTrailingWhitespace03);
  runTest(&euc, testImcaStripTrailingWhitespace04);
  runTest(&euc, testImcaStripTrailingWhitespace05);
  runTest(&euc, testImcaStripTrailingWhitespace06);

  runTest(&euc, testImcaStringEndsWith01);
  runTest(&euc, testImcaStringEndsWith02);
  runTest(&euc, testImcaStringEndsWith03);
  runTest(&euc, testImcaStringEndsWith04);
  runTest(&euc, testImcaStringEndsWith05);
  runTest(&euc, testImcaStringEndsWith06);

  runTest(&euc, testImcaStringEndsWithIgnoringCase01);
  runTest(&euc, testImcaStringEndsWithIgnoringCase02);

  runTest(&euc, testImcaWaitForIpServerConnection01);

  runTest(&euc, testImcaPrintEmailDateField01);
  runTest(&euc, testImcaPrintEmailDateField02);
  runTest(&euc, testImcaPrintEmailFromField01);
  runTest(&euc, testImcaPrintEmailFromField02);
  runTest(&euc, testImcaPrintEmailFromField03);
  runTest(&euc, testImcaPrintEmailToField01);
  runTest(&euc, testImcaPrintEmailToField02);
  runTest(&euc, testImcaPrintEmailToField03);
  runTest(&euc, testImcaPrintEmailToField04);
  runTest(&euc, testImcaPrintEmailToField05);
  runTest(&euc, testImcaPrintEmailToField06);
  runTest(&euc, testImcaPrintEmailToField07);
  runTest(&euc, testImcaPrintEmailToField08);
  runTest(&euc, testImcaPrintEmailToField09);
  runTest(&euc, testImcaPrintEmailToField10);
  runTest(&euc, testImcaPrintEmailToField11);
  runTest(&euc, testImcaPrintEmailCcField01);
  runTest(&euc, testImcaPrintEmailCcField02);
  runTest(&euc, testImcaPrintEmailCcField03);
  runTest(&euc, testImcaPrintEmailSubjectField01);
  runTest(&euc, testImcaPrintEmailSubjectField02);
  runTest(&euc, testImcaPrintEmailSubjectField03);
  runTest(&euc, testImcaPrintEmailBody01);
  runTest(&euc, testImcaPrintEmailBody02);
  runTest(&euc, testImcaPrintEmailBody03);
  runTest(&euc, testImcaPrintEmailBody04);
  runTest(&euc, testImcaPrintEmailBody05);
  runTest(&euc, testImcaCreateEmailMessage01);
  runTest(&euc, testImcaCreateEmailMessage02);
  runTest(&euc, testImcaCreateEmailMessage03);

  runTest(&euc, testImcaIsLastLineOfSmtpResponse01);
  runTest(&euc, testImcaGetStatusFromSmtpResponse01);
  runTest(&euc, testImcaReadSmtpResponse01);
  runTest(&euc, testImcaReadSmtpResponse02);
  runTest(&euc, testImcaWaitForSmtpServerGreeting01);

  runTest(&euc, testImcaSendSmtpHeloRequest01);
  runTest(&euc, testImcaSendSmtpHeloRequest02);
  runTest(&euc, testImcaSendSmtpHeloRequest03);
  runTest(&euc, testImcaSendSmtpHeloRequest04);

  runTest(&euc, testImcaSendSmtpMailRequest01);
  runTest(&euc, testImcaSendSmtpMailRequest02);
  runTest(&euc, testImcaSendSmtpMailRequest03);
  runTest(&euc, testImcaSendSmtpMailRequest04);

  runTest(&euc, testImcaSendSmtpRcptRequests01);
  runTest(&euc, testImcaSendSmtpRcptRequests02);
  runTest(&euc, testImcaSendSmtpRcptRequests03);
  runTest(&euc, testImcaSendSmtpRcptRequests04);
  runTest(&euc, testImcaSendSmtpRcptRequests05);
  runTest(&euc, testImcaSendSmtpRcptRequests06);
  runTest(&euc, testImcaSendSmtpRcptRequests07);
  runTest(&euc, testImcaSendSmtpRcptRequests08);

  runTest(&euc, testImcaSendSmtpDataRequest01);
  runTest(&euc, testImcaSendSmtpDataRequest02);
  runTest(&euc, testImcaSendSmtpDataRequest03);

  runTest(&euc, testImcaSendSmtpEncodedMessage01);
  runTest(&euc, testImcaSendSmtpEncodedMessage02);
  runTest(&euc, testImcaSendSmtpEncodedMessage03);
  runTest(&euc, testImcaSendSmtpEncodedMessage04);
  runTest(&euc, testImcaSendSmtpEncodedMessage05);
  runTest(&euc, testImcaSendSmtpEncodedMessage06);
  runTest(&euc, testImcaSendSmtpEncodedMessage07);
  runTest(&euc, testImcaSendSmtpEncodedMessage08);
  runTest(&euc, testImcaSendSmtpEncodedMessage09);

  runTest(&euc, testImcaSendSmtpQuitRequest01);
  runTest(&euc, testImcaSendSmtpQuitRequest02);

  runTest(&euc, testImcaSendEmail01);
  runTest(&euc, testImcaSendEmail02);
  runTest(&euc, testImcaSendEmail03);

  runTest(&euc, testImcaErrlogTsPrintf01);
}

static void initializeErrlogExpectContext(char *(*expectPatterns)[],
    int isFirstTime) {
  errlogExpectContext.index = 0;
  errlogExpectContext.expectPatterns = expectPatterns;
  errlogExpectContext.hadError = 0;
  if (isFirstTime)
    errlogExpectContext.doneEvent = epicsEventMustCreate(epicsEventEmpty);
}

/*
 * context->expectPatterns must contain "shell" pattern strings followed by a
 * terminating NULL element.
 *
 * context->doneEvent is signaled just before this function returns.
 *
 * context->hadError is set to 1 if there was an error, 0 if the expectations
 * and responses were processed successfully.  This value is only valid after
 * the context->doneEvent has been signaled.
 */
static void errlogExpectListener(void *pvt, const char *message) {
  ImcaErrlogExpectContext *context;
  char *expectedPattern;
  int isLastExpected;

  context = (ImcaErrlogExpectContext *)pvt;
  expectedPattern = (*(context->expectPatterns))[context->index];
  isLastExpected = expectedPattern == NULL ?
    1 : (*(context->expectPatterns))[context->index + 1] == NULL;

  if (expectedPattern == NULL || message == NULL) {
    context->hadError = 1;
  } else {
    if (!epicsStrGlobMatch(message, expectedPattern)) {
      fprintf(stderr, "failure: \"%s\" does not match pattern \"%s\" "
        "(index=%d)\n", message, expectedPattern, context->index);
      context->hadError = 1;
    }
    context->index = context->index + 1;
  }
  if (isLastExpected) epicsEventSignal(context->doneEvent);
}

static asynStatus initializeTestSmtpClientAndServer() {
  asynStatus status;
  epicsEventWaitStatus waitStatus;

  smtpClient = NULL;
  smtpServer = NULL;
  smtpServerConnectionHandler = NULL;

  status = imcaCreateListeningIpServer(&smtpServerConnectionHandler,
    smtpServerAsynPortName, "\n", 1, "", 0);
  if (status != asynSuccess) return status;
  status = imcaCreateConnectedIpClient(&smtpClient, smtpClientAsynPortName,
    "\n", 1, "", 0);
  if (status != asynSuccess) {
    imcaDestroyIpServer(smtpServerConnectionHandler);
    return status;
  }
  waitStatus = imcaWaitForIpServerConnection(&smtpServer,
    smtpServerConnectionHandler, 2.0);
  if (waitStatus != epicsEventWaitOK) {
    imcaDestroyIpClient(smtpClient);
    imcaDestroyIpServer(smtpServerConnectionHandler);
    return asynError;
  }

  return asynSuccess;
}

static void startSmtpExpectAndRespondThread(
    char *(*expectationsAndResponses)[]) {
  initializeSmtpExpectAndRespondContext(0);
  smtpExpectAndRespondContext.expectationsAndResponses =
    expectationsAndResponses;
  epicsThreadCreate("smtpExpectAndRespond", epicsThreadPriorityLow,
    epicsThreadGetStackSize(epicsThreadStackSmall),
    (EPICSTHREADFUNC)smtpExpectAndRespond, &smtpExpectAndRespondContext);
}

/*
 * context->expectationsAndResponses must contain pairs of strings followed by
 * a terminating NULL element.
 *
 * An empty string in context->expectationsAndResponses indicates that a read
 * or write for that expectation or response should not be performed (i.e. a
 * no-op).  To expect an empty line, use the string "\r" since only the "\n"
 * of the end of message sequence "\r\n" will be removed by asyn.
 *
 * context->doneEvent is signaled just before this function returns.
 *
 * context->hadError is set to 1 if there was an error, 0 if the expectations
 * and responses were processed successfully.  This value is only valid after
 * the context->doneEvent has been signaled.
 */
static void smtpExpectAndRespond(ImcaSmtpExpectAndRespondContext *context) {
  int i;
  char *expected;
  char *response;
  /* this array must accommodate largest possible line from test client */
  char actual[IMCA_SMTP_ENCODED_MESSAGE_SEND_BUFFER_SIZE + 10];
  asynStatus status;

  context->hadError = 1;
  for (i = 0; (*(context->expectationsAndResponses))[i] != NULL; i += 2) {
    expected = (*(context->expectationsAndResponses))[i];
    response = (*(context->expectationsAndResponses))[i+1];

    if (strcmp(expected, "") != 0) {
      status = imcaAsynReadString(context->server, actual, sizeof(actual),
        context->timeout);
      if (status != asynSuccess) {
        epicsEventSignal(context->doneEvent);
        return;
      }
      if (strcmp(expected, actual) != 0) {
        epicsEventSignal(context->doneEvent);
        return;
      }
    }

    if (strcmp(response, "") != 0) {
      status = imcaAsynWriteString(context->server, response,
        context->timeout);
      if (status != asynSuccess) {
        epicsEventSignal(context->doneEvent);
        return;
      }
    }
  }

  context->hadError = 0;
  epicsEventSignal(context->doneEvent);
  return;
}

static asynStatus destroyTestSmtpClientAndServer() {
  asynStatus status;
  asynStatus status2;
  asynStatus status3;

  status = imcaDestroyIpClient(smtpClient);
  status2 = imcaDestroyIpServerSession(smtpServer);
  status3 = imcaDestroyIpServer(smtpServerConnectionHandler);

  smtpClient = NULL;
  smtpServer = NULL;
  smtpServerConnectionHandler = NULL;

  if (status != asynSuccess) return status;
  if (status2 != asynSuccess) return status2;
  if (status3 != asynSuccess) return status3;
  return asynSuccess;
}

static void initializeSmtpClientAndServerInfo(
    const char *inSmtpClientAsynPortName,
    const char *inSmtpServerAsynPortName) {
  assert(strlen(inSmtpClientAsynPortName) < sizeof(smtpClientAsynPortName));
  strncpy(smtpClientAsynPortName, inSmtpClientAsynPortName,
    sizeof(smtpClientAsynPortName) - 1);
  smtpClientAsynPortName[sizeof(smtpClientAsynPortName) - 1] = '\0';

  assert(strlen(inSmtpServerAsynPortName) < sizeof(smtpServerAsynPortName));
  strncpy(smtpServerAsynPortName, inSmtpServerAsynPortName,
    sizeof(smtpServerAsynPortName) - 1);
  smtpServerAsynPortName[sizeof(smtpServerAsynPortName) - 1] = '\0';
}

static void initializeSmtpExpectAndRespondContext(int isFirstTime) {
  smtpExpectAndRespondContext.expectationsAndResponses = NULL;
  smtpExpectAndRespondContext.server = smtpServer;
  smtpExpectAndRespondContext.timeout = 5.0;
  smtpExpectAndRespondContext.hadError = 0;
  if (isFirstTime)
    smtpExpectAndRespondContext.doneEvent =
      epicsEventMustCreate(epicsEventEmpty);
}

void imcalibTestRunner(const char *inSmtpClientAsynPortName,
    const char *inSmtpServerAsynPortName) {
  initializeErrlogExpectContext(NULL, 1);
  initializeSmtpClientAndServerInfo(inSmtpClientAsynPortName,
    inSmtpServerAsynPortName);
  initializeSmtpExpectAndRespondContext(1);
  imcaInitializeEpicsUnit(&euc);
  imcaPrintEpicsUnitTestsStarting();
  imcaNoteNowAsEpicsUnitTestsStartTime(&euc);
  runAllTests();
  imcaNoteNowAsEpicsUnitTestsEndTime(&euc);
  imcaPrintEpicsUnitTestsSummary(&euc);
}

static const iocshArg imcalibTestRunnerArg0 = { "smtpClientAsynPortName",
  iocshArgString };
static const iocshArg imcalibTestRunnerArg1 = { "smtpServerAsynPortName",
  iocshArgString };
static const iocshArg *const imcalibTestRunnerArgs[] = {
  &imcalibTestRunnerArg0, &imcalibTestRunnerArg1 };
static const iocshFuncDef imcalibTestRunnerDef = {
  "imcalibTestRunner", 2, imcalibTestRunnerArgs };
static void imcalibTestRunnerCall(const iocshArgBuf *args) {
  imcalibTestRunner(args[0].sval, args[1].sval);
}
static void imcalibTestRegistrar(void) {
  static int isFirstTime = 1;
  if (!isFirstTime) return;
  isFirstTime = 0;
  iocshRegister(&imcalibTestRunnerDef, imcalibTestRunnerCall);
}
epicsExportRegistrar(imcalibTestRegistrar);
